from pytest import fixture, mark
import numpy as np
from numpy.testing import assert_equal, assert_allclose

import quimb as qu
from quimb.linalg.numpy_linalg import (
    sort_inds,
    eigs_numpy,
)


@fixture
def xs():
    return np.array([-2.4 - 1j, -1 + 2.2j, 1 - 2.1j, 2.3 + 1j])


@fixture
def ham1():
    evecs = qu.rand_uni(5)
    evals = np.array([-5, -3, 0.1, 2, 4])
    return qu.dot(evecs, qu.ldmul(evals, evecs.H))


class TestSortInds:
    @mark.parametrize("method, inds, sigma",
                      [("LM", [0, 3, 1, 2], None),
                       ("SM", [2, 1, 3, 0], None),
                       ("SA", [0, 1, 2, 3], None),
                       ("SR", [0, 1, 2, 3], None),
                       ("SI", [2, 0, 3, 1], None),
                       ("LA", [3, 2, 1, 0], None),
                       ("LR", [3, 2, 1, 0], None),
                       ("LI", [1, 3, 0, 2], None),
                       ("TM", [1, 2, 3, 0], 2.41),
                       ("tm", [1, 2, 3, 0], 2.41),
                       ("TR", [2, 3, 1, 0], 1.01),
                       ("TI", [3, 1, 0, 2], 1.01)])
    def test_simple(self, xs, method, inds, sigma):
        assert_equal(sort_inds(xs, method, sigma), inds)


class TestNumpyEigk:
    @mark.parametrize("which, k, ls, sigma",
                      [("lm", 3, [-5, 4, -3], None),
                       ("sm", 3, [0.1, 2, -3], None),
                       ("tm", 3, [-3, 2, 4], 2.9)])
    def test_evals(self, ham1, which, k, ls, sigma):
        lk = eigs_numpy(ham1, k=k, which=which, return_vecs=False,
                        sigma=sigma, sort=False)
        assert_allclose(lk, ls)

    @mark.parametrize("which, k, sigma",
                      [("sa", 5, None)])
    def test_evecs(self, ham1, which, k, sigma):
        lk, vk = eigs_numpy(ham1, k=k, which=which, return_vecs=True,
                            sigma=sigma, sort=False)
        assert isinstance(vk, qu.qarray)
        assert_allclose(qu.dot(vk, qu.ldmul(lk, vk.H)), ham1)


class TestAutoBlock:

    def test_eigh(self):
        H = qu.ham_mbl(6, dh=2.5)
        a_el, a_ev = qu.eigh(H, autoblock=False)
        el, ev = qu.eigh(H, autoblock=True)

        assert qu.norm(ev @ qu.ldmul(el, ev.H) - H, 'fro') < 1e-12
        assert_allclose(a_el, el)
        assert_allclose(ev.H @ ev, np.eye(H.shape[0]), atol=1e-12)

    def test_eigvals(self):
        H = qu.ham_hubbard_hardcore(4)
        a_el = qu.eigvalsh(H, autoblock=False)
        el = qu.eigvalsh(H, autoblock=True)
        assert_allclose(a_el, el, atol=1e-12)



import pytest
from pytest import approx
import numpy as np

import quimb as qu
import quimb.tensor as qtn
from quimb.tensor.tensor_1d_tebd import OTOC_local


class TestTEBD:

    def test_setup_and_sweep(self):
        n = 10
        H_int = qu.ham_heis(n=2, cyclic=False)
        psi0 = qtn.MPS_neel_state(n, dtype=complex)
        tebd = qtn.TEBD(psi0, H_int, dt=0.05)
        assert tebd.pt.bond_size(0, 1) == 1
        tebd.sweep('right', 1 / 2)
        assert tebd.pt.count_canonized() == (n - 1, 0)
        tebd.sweep('left', 1 / 2)
        assert tebd.pt.count_canonized() == (0, n - 1)
        assert tebd.pt.bond_size(0, 1) > 1
        assert not tebd._queued_sweep

    @pytest.mark.parametrize('cyclic', [False, True])
    @pytest.mark.parametrize('order', [2, 4])
    @pytest.mark.parametrize('dt,tol', [
        (0.0759283, None),
        (None, 1e-4),
        (0.0759283, 1e-4),
        (None, None),
    ])
    @pytest.mark.parametrize('n', [5, 6])
    def test_evolve_obc_pbc(self, n, order, dt, tol, cyclic):
        tf = 1.0 if cyclic else 2
        psi0 = qtn.MPS_neel_state(n, cyclic=cyclic)
        H_int = qu.ham_heis(2, cyclic=False)  # this is just the interaction

        if dt and tol:
            with pytest.raises(ValueError):
                qtn.TEBD(psi0, H_int, dt=dt, tol=tol)
            return

        tebd = qtn.TEBD(psi0, H_int, dt=dt, tol=tol)
        assert tebd.cyclic == cyclic
        tebd.split_opts['cutoff'] = 1e-10

        if (dt is None and tol is None):
            with pytest.raises(ValueError):
                tebd.update_to(tf, order=order)
            return

        tebd.update_to(tf, order=order)
        assert tebd.t == approx(tf)
        assert not tebd._queued_sweep

        dpsi0 = psi0.to_dense()
        dham = qu.ham_heis(n=n, sparse=True, cyclic=cyclic)
        evo = qu.Evolution(dpsi0, dham)
        evo.update_to(tf)

        assert qu.expec(evo.pt, tebd.pt.to_dense()) == approx(1, rel=1e-3 if
                                                              cyclic else 1e-5)

    @pytest.mark.parametrize('cyclic', [False, True])
    @pytest.mark.parametrize('order', [2, 4])
    @pytest.mark.parametrize('dt,tol', [
        (0.0759283, None),
        (None, 1e-4),
        (0.0759283, 1e-4),
        (None, None),
    ])
    @pytest.mark.parametrize('n', [5, 6])
    def test_imag_evolve_obc_pbc(self, n, order, dt, tol, cyclic):
        tf = 2
        ground = qtn.MPS_computational_state('0' * n, cyclic=cyclic)
        excited = qtn.MPS_computational_state(('01' * n)[:n], cyclic=cyclic)
        psi0 = (ground + excited) / 2**0.5
        H = qtn.ham_1d_ising(n, j=-1, cyclic=cyclic)

        if dt and tol:
            with pytest.raises(ValueError):
                qtn.TEBD(psi0, H, dt=dt, tol=tol, imag=True)
            return

        tebd = qtn.TEBD(psi0, H, dt=dt, tol=tol, imag=True)
        assert tebd.cyclic == cyclic
        tebd.split_opts['cutoff'] = 1e-10

        if (dt is None and tol is None):
            with pytest.raises(ValueError):
                tebd.update_to(tf, order=order)
            return

        tebd.update_to(tf, order=order)
        assert tebd.t == approx(tf)
        assert not tebd._queued_sweep

        H_mpo = qtn.MPO_ham_ising(n, j=-1, cyclic=cyclic)
        E_ground = qtn.expec_TN_1D(ground.H, H_mpo, ground)
        E_excited = qtn.expec_TN_1D(excited.H, H_mpo, excited)

        psi1 = (np.exp(-tf * E_ground) * ground +
                np.exp(-tf * E_excited) * excited)
        psi1 /= np.sqrt(np.exp(-2 * tf * E_ground) +
                        np.exp(-2 * tf * E_excited))

        assert qtn.expec_TN_1D(psi1.H, tebd.pt) == approx(1, rel=1e-5)

    @pytest.mark.parametrize('cyclic', [False, True])
    @pytest.mark.parametrize('dt,tol', [
        (0.0659283, None),
        (None, 1e-5),
    ])
    def test_at_times(self, dt, tol, cyclic):
        n = 10
        psi0 = qtn.MPS_neel_state(n, cyclic=cyclic)
        H_int = qu.ham_heis(2, cyclic=False)
        tebd = qtn.TEBD(psi0, H_int, dt=dt, tol=tol)

        for pt in tebd.at_times([0.1, 0.2, 0.3, 0.4, 0.5]):
            assert pt.H @ pt == approx(1, rel=1e-5)

        assert tebd.err <= 1e-5

    def test_local_ham_1d_and_single_site_terms(self):
        n = 10
        psi0 = qtn.MPS_neel_state(n)
        lham_1d = qtn.ham_1d_XY(n, bz=0.9)
        tebd = qtn.TEBD(psi0, lham_1d)
        tebd.update_to(1.0, tol=1e-5)
        assert abs(psi0.H @ tebd.pt) < 1.0
        assert tebd.pt.entropy(5) > 0.0

        psi0_dns = qu.neel_state(n)
        H_dns = qu.ham_XY(10, jxy=1.0, bz=0.9, cyclic=False)
        evo = qu.Evolution(psi0_dns, H_dns)
        evo.update_to(1.0)

        assert qu.expec(tebd.pt.to_dense(), evo.pt) == pytest.approx(1.0)

    def test_local_ham_1d_and_single_site_terms_heis(self):
        n = 10
        psi0 = qtn.MPS_neel_state(n)
        lham_1d = qtn.ham_1d_heis(n, j=(0.7, 0.8, 0.9), bz=0.337)
        tebd = qtn.TEBD(psi0, lham_1d)
        tebd.update_to(1.0, tol=1e-5)
        assert abs(psi0.H @ tebd.pt) < 1.0
        assert tebd.pt.entropy(5) > 0.0

        psi0_dns = qu.neel_state(n)
        H_dns = qu.ham_heis(10, j=(0.7, 0.8, 0.9), b=0.337, cyclic=False)
        evo = qu.Evolution(psi0_dns, H_dns)
        evo.update_to(1.0)

        assert qu.expec(tebd.pt.to_dense(), evo.pt) == pytest.approx(1.0)

    def test_non_trans_invar(self):
        n = 10
        tf = 1.0
        p0 = qtn.MPS_rand_state(n, bond_dim=1)
        H = qtn.ham_1d_mbl(n, dh=1.7, cyclic=False, seed=42)
        print(H)
        tebd = qtn.TEBD(p0, H)
        tebd.update_to(tf, tol=1e-3)

        p0d = p0.to_dense()
        Hd = qu.ham_mbl(n, dh=1.7, cyclic=False, seed=42, sparse=True)
        evo = qu.Evolution(p0d, Hd)
        evo.update_to(tf)

        assert qu.expec(tebd.pt.to_dense(), evo.pt) == pytest.approx(1.0)

    @pytest.mark.parametrize('cyclic', [False, True])
    def test_ising_model_with_field(self, cyclic):

        p = qtn.MPS_computational_state('0000100000', cyclic=cyclic)
        pd = p.to_dense()

        lham_1d = qtn.ham_1d_ising(10, j=4, bx=1, cyclic=cyclic)
        H_mpo = qtn.MPO_ham_ising(10, j=4, bx=1, cyclic=cyclic)
        H = qu.ham_ising(10, jz=4, bx=1, cyclic=cyclic)

        tebd = qtn.TEBD(p, lham_1d, tol=1e-6)
        tebd.split_opts['cutoff'] = 1e-9
        tebd.split_opts['cutoff_mode'] = 'rel'
        evo = qu.Evolution(pd, H)

        e0 = qu.expec(pd, H)
        e0_mpo = qtn.expec_TN_1D(p.H, H_mpo, p)

        assert e0_mpo == pytest.approx(e0)

        tf = 2
        ts = np.linspace(0, tf, 21)
        evo.update_to(tf)

        for pt in tebd.at_times(ts):
            assert isinstance(pt, qtn.MatrixProductState)
            assert (pt.H @ pt) == pytest.approx(1.0, rel=1e-5)

        assert (qu.expec(tebd.pt.to_dense(), evo.pt) ==
                pytest.approx(1.0, rel=1e-5))

        ef_mpo = qtn.expec_TN_1D(tebd.pt.H, H_mpo, tebd.pt)
        assert ef_mpo == pytest.approx(e0, 1e-5)


def test_OTOC_local():
    L = 10
    psi0 = qtn.MPS_computational_state('0' * L, cyclic=True)
    H1 = qtn.ham_1d_ising(L, j=4, bx=0, cyclic=True)
    H_back1 = qtn.ham_1d_ising(L, j=-4, bx=0, cyclic=True)
    H2 = qtn.ham_1d_ising(L, j=4, bx=1, cyclic=True)
    H_back2 = qtn.ham_1d_ising(L, j=-4, bx=-1, cyclic=True)
    A = qu.pauli('z')
    ts = np.linspace(1, 2, 2)
    OTOC_t = []
    for OTOC in OTOC_local(psi0, H1, H_back1, ts, 5, A, tol=1e-5,
                           split_opts={'cutoff': 1e-5, 'cutoff_mode': 'rel'},
                           initial_eigenstate='check'):
        OTOC_t += [OTOC]
    assert OTOC_t[0] == pytest.approx(1.0)
    assert OTOC_t[1] == pytest.approx(1.0)
    x_t = []
    for x in OTOC_local(psi0, H2, H_back2, ts, 5, A, tol=1e-5,
                        split_opts={'cutoff': 1e-5, 'cutoff_mode': 'rel'},
                        initial_eigenstate='check'):
        x_t += [x]
    assert x_t[0] == pytest.approx(0.52745, rel=1e-4, abs=1e-9)
    assert x_t[1] == pytest.approx(0.70440, rel=1e-4, abs=1e-9)



import itertools


def sort_unique(edges):
    """Make sure there are no duplicate edges and that for each
    ``coo_a < coo_b``.
    """
    return tuple(sorted(
        tuple(sorted(edge))
        for edge in set(map(frozenset, edges))
    ))


# ----------------------------------- 2D ------------------------------------ #

def check_2d(coo, Lx, Ly, cyclic):
    """Check ``coo`` in inbounds for a maybe cyclic 2D lattice.
    """
    x, y = coo
    if (not cyclic) and not ((0 <= x < Lx) and (0 <= y < Ly)):
        return
    return (x % Lx, y % Ly)


def edges_2d_square(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D square lattice. The nodes
    (sites) are labelled like ``(i, j)``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int), (int, int))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for coob in [(i, j + 1), (i + 1, j)]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j), coob))

    return sort_unique(edges)


def edges_2d_hexagonal(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D hexagonal lattice. There are two
    sites per cell, and note the cells do not form a square tiling. The nodes
    (sites) are labelled like ``(i, j, s)`` for ``s`` in ``'AB'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int, str), (int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for *coob, lbl in [
            (i, j, 'B'),
            (i, j - 1, 'B'),
            (i - 1, j, 'B'),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'A'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, 'A'),
            (i, j + 1, 'A'),
            (i + 1, j, 'A'),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'B'), (*coob, lbl)))

    return sort_unique(edges)


def edges_2d_triangular(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D triangular lattice. There is a
    single site per cell, and note the cells do not form a square tiling.
    The nodes (sites) are labelled like ``(i, j)``.

    Parameters
    ----------
    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int), (int, int))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for coob in [(i, j + 1), (i + 1, j), (i + 1, j - 1)]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j), coob))

    return sort_unique(edges)


def edges_2d_triangular_rectangular(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D triangular lattice tiled in a
    rectangular geometry. There are two sites per rectangular cell. The nodes
    (sites) are labelled like ``(i, j, s)`` for ``s`` in ``'AB'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int, s), (int, int, s))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for *coob, lbl in [
            (i, j, 'B'),
            (i, j - 1, 'B'),
            (i, j + 1, 'A'),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'A'), (*coob, lbl)))

        for *coob, lbl in [
            (i + 1, j, 'A'),
            (i, j + 1, 'B'),
            (i + 1, j + 1, 'A'),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'B'), (*coob, lbl)))

    return sort_unique(edges)


def edges_2d_kagome(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D kagome lattice. There are
    three sites per cell, and note the cells do not form a square tiling. The
    nodes (sites) are labelled like ``(i, j, s)`` for ``s`` in ``'ABC'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int, str), (int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for *coob, lbl in [
            (i, j, 'B'),
            (i, j - 1, 'B'),
            (i, j, 'C'),
            (i - 1, j, 'C')
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'A'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, 'C'),
            (i - 1, j + 1, 'C'),
            (i, j, 'A'),
            (i, j + 1, 'A')
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'B'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, 'A'),
            (i + 1, j, 'A'),
            (i, j, 'B'),
            (i + 1, j - 1, 'B')
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, 'C'), (*coob, lbl)))

    return sort_unique(edges)


# ----------------------------------- 3D ------------------------------------ #

def check_3d(coo, Lx, Ly, Lz, cyclic):
    """Check ``coo`` in inbounds for a maybe cyclic 3D lattice.
    """
    x, y, z = coo
    OBC = not cyclic
    inbounds = (0 <= x < Lx) and (0 <= y < Ly) and (0 <= z < Lz)
    if OBC and not inbounds:
        return
    return (x % Lx, y % Ly, z % Lz)


def edges_3d_cubic(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D cubic lattice. The nodes
    (sites) are labelled like ``(i, j, k)``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int), (int, int, int))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for coob in [(i, j, k + 1), (i, j + 1, k), (i + 1, j, k)]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k), coob))

    return sort_unique(edges)


def edges_3d_pyrochlore(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D pyorchlore lattice. There are
    four sites per cell, and note the cells do not form a cubic tiling. The
    nodes (sites) are labelled like ``(i, j, k, s)`` for ``s`` in ``'ABCD'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int, str), (int, int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for *coob, lbl in [
            (i, j, k, 'B'),
            (i, j - 1, k, 'B'),
            (i, j, k, 'C'),
            (i - 1, j, k, 'C'),
            (i, j, k, 'D'),
            (i, j, k - 1, 'D'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'A'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, 'C'),
            (i - 1, j + 1, k, 'C'),
            (i, j, k, 'D'),
            (i, j + 1, k - 1, 'D'),
            (i, j, k, 'A'),
            (i, j + 1, k, 'A'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'B'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, 'D'),
            (i + 1, j, k - 1, 'D'),
            (i, j, k, 'A'),
            (i + 1, j, k, 'A'),
            (i, j, k, 'B'),
            (i + 1, j - 1, k, 'B'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'C'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, 'A'),
            (i, j, k + 1, 'A'),
            (i, j, k, 'B'),
            (i, j - 1, k + 1, 'B'),
            (i, j, k, 'C'),
            (i - 1, j, k + 1, 'C'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'D'), (*coob, lbl)))

    return sort_unique(edges)


def edges_3d_diamond(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D diamond lattice. There are
    two sites per cell, and note the cells do not form a cubic tiling.  The
    nodes (sites) are labelled like ``(i, j, k, s)`` for ``s`` in ``'AB'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int, str), (int, int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for *coob, lbl in [
            (i, j, k, 'B'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'A'), (*coob, lbl)))

        for *coob, lbl in [
                (i, j, k + 1, 'A'),
                (i, j + 1, k, 'A'),
                (i + 1, j, k, 'A'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'B'), (*coob, lbl)))

    return sort_unique(edges)


def edges_3d_diamond_cubic(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D diamond lattice tiled in a cubic
    geometry. There are eight sites per cubic cell. The nodes (sites) are
    labelled like ``(i, j, k, s)`` for ``s`` in ``'ABCDEFGH'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int, str), (int, int, int, str))]
    """

    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for *coob, lbl in [
            (i, j, k, 'E'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'A'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, 'E'),
            (i, j, k, 'F'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'B'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, 'E'),
            (i, j, k, 'G'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'C'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, 'E'),
            (i, j, k, 'H'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'D'), (*coob, lbl)))

        for *coob, lbl in []:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'E'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j + 1, k, 'C'),
            (i + 1, j, k, 'D'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'F'), (*coob, lbl)))

        for *coob, lbl in [
            (i + 1, j, k + 1, 'A'),
            (i, j, k + 1, 'B'),
            (i + 1, j, k, 'D'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'G'), (*coob, lbl)))

        for *coob, lbl in [
            (i, j + 1, k + 1, 'A'),
            (i, j, k + 1, 'B'),
            (i, j + 1, k, 'C'),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, 'H'), (*coob, lbl)))

    return sort_unique(edges)



"""Decorator for automatically just in time compiling tensor network functions.

TODO:

    - [ ] go via an intermediate pytree / array function, that could be shared
          e.g. with the TNOptimizer class.

"""
import functools

import autoray as ar



class AutojittedTN:
    """Class to hold the ``autojit_tn`` decorated function callable.
    """

    def __init__(
        self,
        fn,
        decorator=ar.autojit,
        check_inputs=True,
        **decorator_opts
    ):
        self.fn = fn
        self.fn_store = {}
        self.decorator_opts = decorator_opts
        self.check_inputs = check_inputs
        self.decorator = decorator

    def setup_fn(self, tn, *args, **kwargs):
        from quimb.tensor import TensorNetwork

        @self.decorator(**self.decorator_opts)
        def fn_jit(arrays):
            # use separate TN to trace through function
            jtn = tn.copy()

            # insert the tracing arrays
            for t, array in zip(jtn, arrays):
                t.modify(data=array)

            # run function on TN with tracing arrays
            result = self.fn(jtn, *args, **kwargs)

            # check for a inplace tn function
            if isinstance(result, TensorNetwork):
                if result is not jtn:
                    raise ValueError(
                        "If you are compiling a function that returns a"
                        " tensor network it needs to be inplace.")
                self.inplace = True
                return tuple(t.data for t in jtn)
            else:
                # function returns raw scalar/array(s)
                self.inplace = False
                return result

        return fn_jit

    def __call__(self, tn, *args, **kwargs):

        # do we need to generate a new function for these inputs
        if self.check_inputs:
            key = (
                tn.geometry_hash(strict_index_order=True),
                tuple(args),
                tuple(sorted(kwargs.items())),
            )
        else:
            # always use the same function
            key = None

        if key not in self.fn_store:
            self.fn_store[key] = self.setup_fn(tn, *args, **kwargs)
        fn_jit = self.fn_store[key]

        # run the compiled function
        arrays = tuple(t.data for t in tn)
        out = fn_jit(arrays)

        if self.inplace:
            # reinsert output arrays into input TN structure
            for t, array in zip(tn, out):
                t.modify(data=array)
            return tn

        return out


def autojit_tn(
    fn=None,
    decorator=ar.autojit,
    check_inputs=True,
    **decorator_opts,
):
    """Decorate a tensor network function to be just in time compiled / traced.
    This traces solely array operations resulting in a completely static
    computational graph with no side-effects. The resulting function can be
    much faster if called repeatedly with only numeric changes, or hardware
    accelerated if a library such as ``jax`` is used.

    Parameters
    ----------
    fn : callable
        The function to be decorated. It should take as its first argument a
        :class:`~quimb.tensor.tensor_core.TensorNetwork` and return either act
        inplace on it or return a raw scalar or array.
    decorator : callable
        The decorator to use to wrap the underlying array function. For example
        ``jax.jit``. Defaults to ``autoray.autojit``.
    check_inputs : bool, optional
        Whether to check the inputs to the function every call to see if a new
        compiled function needs to be generated. If ``False`` the same compiled
        function will be used for all inputs which might be incorrect. Defaults
        to ``True``.
    decorator_opts
        Options to pass to the decorator, e.g. ``backend`` for
        ``autoray.autojit``.
    """
    kwargs = {
        'decorator': decorator,
        'check_inputs': check_inputs,
        **decorator_opts,
    }
    if fn is None:
        return functools.partial(autojit_tn, **kwargs)
    return functools.wraps(fn)(AutojittedTN(fn, **kwargs))



"""Functionailty for drawing tensor networks.
"""
import textwrap
import importlib
import collections

import numpy as np

from ..utils import valmap


HAS_FA2 = importlib.util.find_spec('fa2') is not None


def parse_dict_to_tids_or_inds(spec, tn, default='__NONE__'):
    """Parse a dictionary possibly containing a mix of tags, tids and inds, to
    a dictionary with only sinlge tids and inds as keys. If a tag or set of
    tags are given as a key, all matching tensor tids will receive the value.
    """
    #
    if (spec is not None) and (not isinstance(spec, dict)):
        # assume new default value for everything
        return collections.defaultdict(lambda: spec)

    # allow not specifying a default value
    if default != '__NONE__':
        new = collections.defaultdict(lambda: default)
    else:
        new = {}

    if spec is None:
        return new

    # parse the special values
    for k, v in spec.items():
        if (
            # given as tid
            (isinstance(k, int) and k in tn.tensor_map) or
            # given as ind
            (isinstance(k, str) and k in tn.ind_map)
        ):
            # already a tid
            new[k] = v
            continue

        try:
            for tid in tn._get_tids_from_tags(k):
                new[tid] = v
        except KeyError:
            # just ignore keys that don't match any tensor
            pass

    return new


def _add_or_merge_edge(G, u, v, attrs):
    if not G.has_edge(u, v):
        G.add_edge(u, v, **attrs)
    else:
        # multibond - update attrs
        attrs0 = G.edges[u, v]
        # average colors
        attrs0['color'] = tuple(
            (x + y) / 2 for x, y in zip(attrs0['color'], attrs['color']))
        attrs0['ind'] += ' ' + attrs['ind']
        # hide original edge and instead track multiple bond sizes
        attrs0['multiedge_inds'].append(attrs['ind'])
        attrs0['multiedge_sizes'].append(attrs['edge_size'])
        attrs0['spring_weight'] /= (attrs['edge_size'] + 1)
        attrs0['edge_size'] = 0


def draw_tn(
    tn,
    color=None,
    *,
    output_inds=None,
    highlight_inds=(),
    highlight_tids=(),
    highlight_inds_color=(1.0, 0.2, 0.2, 1.0),
    highlight_tids_color=(1.0, 0.2, 0.2, 1.0),
    show_inds=None,
    show_tags=None,
    show_scalars=True,
    custom_colors=None,
    title=None,
    legend=True,
    fix=None,
    k=None,
    iterations='auto',
    initial_layout='auto',
    use_forceatlas2=1000,
    use_spring_weight=False,
    node_color=None,
    node_scale=1.0,
    node_size=None,
    node_alpha=1.0,
    node_shape='o',
    node_outline_size=None,
    node_outline_darkness=0.8,
    node_hatch='',
    edge_color=None,
    edge_scale=1.0,
    edge_alpha=1 / 2,
    multiedge_spread=0.1,
    show_left_inds=True,
    arrow_closeness=1.1,
    arrow_length=1.0,
    arrow_overhang=1.0,
    arrow_linewidth=1.0,
    label_color=None,
    font_size=10,
    font_size_inner=7,
    font_family='monospace',
    isdark=None,
    figsize=(6, 6),
    margin=None,
    xlims=None,
    ylims=None,
    get=None,
    return_fig=False,
    ax=None,
):
    """Plot this tensor network as a networkx graph using matplotlib,
    with edge width corresponding to bond dimension.

    Parameters
    ----------
    color : sequence of tags, optional
        If given, uniquely color any tensors which have each of the tags.
        If some tensors have more than of the tags, only one color will show.
    output_inds : sequence of str, optional
        For hyper tensor networks explicitly specify which indices should be
        drawn as outer indices. If not set, the outer indices are assumed to be
        those that only appear on a single tensor.
    highlight_inds : iterable, optional
        Highlight these edges.
    highlight_tids : iterable, optional
        Highlight these nodes.
    highlight_inds_color
        What color to use for ``highlight_inds`` nodes.
    highlight_tids_color : tuple[float], optional
        What color to use for ``highlight_tids`` nodes.
    show_inds : {None, False, True, 'all', 'bond-size'}, optional
        Explicitly turn on labels for each tensors indices.
    show_tags : {None, False, True}, optional
        Explicitly turn on labels for each tensors tags.
    show_scalars : bool, optional
        Whether to show scalar tensors (floating nodes with no edges).
    custom_colors : sequence of colors, optional
        Supply a custom sequence of colors to match the tags given
        in ``color``.
    title : str, optional
        Set a title for the axis.
    legend : bool, optional
        Whether to draw a legend for the colored tags.
    fix : dict[tags_ind_or_tid], (float, float)], optional
        Used to specify actual relative positions for each tensor node.
        Each key should be a sequence of tags that uniquely identifies a
        tensor, a ``tid``, or a ``ind``, and each value should be a ``(x, y)``
        coordinate tuple.
    k : float, optional
        The optimal distance between nodes.
    iterations : int, optional
        How many iterations to perform when when finding the best layout
        using node repulsion. Ramp this up if the graph is drawing messily.
    initial_layout : {'auto', 'spectral', 'kamada_kawai', 'circular', \\
                      'planar', 'random', 'shell', 'bipartite', ...}, optional
        The name of a networkx layout to use before iterating with the
        spring layout. Set ``iterations=0`` if you just want to use this
        layout only.
    use_forceatlas2 : bool or int, optional
        Whether to try and use ``forceatlas2`` (``fa2``) for the spring layout
        relaxation instead of ``networkx``. If an integer, only try and use
        beyond that many nodes (it can give messier results on smaller graphs).
    use_spring_weight : bool, optional
        Whether to use inverse bond sizes as spring weights to the force
        repulsion layout algorithms.
    node_color : tuple[float], optional
        Default color of nodes.
    node_scale : float, optional
        Scale the node sizes by this factor, in addition to the automatic
        scaling based on the number of tensors.
    node_size : None, float or dict, optional
        How big to draw the tensors. Can be a global single value, or a dict
        containing values for specific tags or tids. This is in absolute
        figure units. See ``node_scale`` simply scale the node sizes up or
        down.
    node_alpha : float, optional
        Transparency of the nodes.
    node_shape : None, str or dict, optional
        What shape to draw the tensors. Should correspond to a matplotlib
        scatter marker. Can be a global single value, or a dict containing
        values for specific tags or tids.
    node_outline_size : None, float or dict, optional
        The width of the border of each node. Can be a global single value, or
        a dict containing values for specific tags or tids.
    node_outline_darkness : float, optional
        Darkening of nodes outlines.
    edge_color : tuple[float], optional
        Default color of edges.
    edge_scale : float, optional
        How much to scale the width of the edges.
    edge_alpha : float, optional
        Set the alpha (opacity) of the drawn edges.
    multiedge_spread : float, optional
        How much to spread the lines of multi-edges.
    show_left_inds : bool, optional
        Whether to show ``tensor.left_inds`` as incoming arrows.
    arrow_closeness : float, optional
        How close to draw the arrow to its target.
    arrow_length : float, optional
        The size of the arrow with respect to the edge.
    arrow_overhang : float, optional
        Varies the arrowhead between a triangle (0.0) and 'V' (1.0).
    label_color : tuple[float], optional
        Color to draw labels with.
    font_size : int, optional
        Font size for drawing tags and outer indices.
    font_size_inner : int, optional
        Font size for drawing inner indices.
    font_family : str, optional
        Font family to use for all labels.
    isdark : bool, optional
        Explicitly specify that the background is dark, and use slightly
        different default drawing colors. If not specified detects
        automatically from `matplotlib.rcParams`.
    figsize : tuple of int, optional
        The size of the drawing.
    margin : None or float, optional
        Specify an argument for ``ax.margin``, else the plot limits will try
        and be computed based on the node positions and node sizes.
    xlims : None or tuple, optional
        Explicitly set the x plot range.
    xlims : None or tuple, optional
        Explicitly set the y plot range.
    get : {None, 'pos', 'graph'}, optional
        If ``None`` then plot as normal, else if:

            - ``'pos'``, return the plotting positions of each ``tid`` and
              ``ind`` drawn as a node, this can supplied to subsequent calls as
              ``fix=pos`` to maintain positions, even as the graph structure
              changes.
            - ``'graph'``, return the ``networkx.Graph`` object. Note that this
              will potentially have extra nodes representing output and hyper
              indices.

    return_fig : bool, optional
        If True and ``ax is None`` then return the figure created rather than
        executing ``pyplot.show()``.
    ax : matplotlib.Axis, optional
        Draw the graph on this axis rather than creating a new figure.
    """
    import networkx as nx
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    from matplotlib.colors import to_rgb, to_rgba
    import math

    if output_inds is None:
        output_inds = set(tn.outer_inds())
    elif isinstance(output_inds, str):
        output_inds = {output_inds}
    else:
        output_inds = set(output_inds)

    # automatically decide whether to show tags and inds
    if show_inds is None:
        show_inds = (len(tn.outer_inds()) <= 20)
    if show_tags is None:
        show_tags = (len(tn.tag_map) <= 20)

    if isdark is None:
        isdark = sum(to_rgb(mpl.rcParams['figure.facecolor'])) / 3 < 0.5

    if isdark:
        default_draw_color = (0.55, 0.57, 0.60, 1.0)
        default_label_color = (.85, .85, .85, 1.0)
    else:
        default_draw_color = (0.45, 0.47, 0.50, 1.0)
        default_label_color = (.25, .25, .25, 1.0)

    if edge_color is None:
        edge_color = default_draw_color
    else:
        edge_color = mpl.colors.to_rgb(edge_color)

    if node_color is None:
        node_color = default_draw_color
    else:
        node_color = mpl.colors.to_rgb(node_color)

    if label_color is None:
        label_color = default_label_color
    elif label_color == 'inherit':
        label_color = mpl.rcParams['axes.labelcolor']

    highlight_tids_color = to_rgba(highlight_tids_color)
    highlight_inds_color = to_rgba(highlight_inds_color)

    # set the size of the nodes and their border
    node_size = parse_dict_to_tids_or_inds(
        node_size, tn,
        default=node_scale * 1000 / tn.num_tensors**0.7)
    node_outline_size = parse_dict_to_tids_or_inds(
        node_outline_size, tn,
        default=min(3, node_size.default_factory()**0.5 / 5))
    node_shape = parse_dict_to_tids_or_inds(
        node_shape, tn, default='o')
    node_hatch = parse_dict_to_tids_or_inds(
        node_hatch, tn, default='')

    # build the graph
    G = nx.Graph()
    hyperedges = []
    node_labels = dict()
    edge_labels = dict()

    for ix, tids in tn.ind_map.items():
        # general information for this index
        edge_attrs = {
            'color': (highlight_inds_color if ix in highlight_inds else
                      edge_color),
            'ind': ix,
            'edge_size': edge_scale * math.log2(tn.ind_size(ix)),
        }
        edge_attrs['multiedge_inds'] = [edge_attrs['ind']]
        edge_attrs['multiedge_sizes'] = [edge_attrs['edge_size']]
        edge_attrs['spring_weight'] = 1 / sum(t.ndim for t in tn._inds_get(ix))

        if (ix in output_inds) or (len(tids) != 2):
            # hyper or outer edge - needs dummy 'node' shown with zero size
            hyperedges.append(ix)
            for tid in tids:
                _add_or_merge_edge(G, tid, ix, edge_attrs)
        else:
            # standard edge
            _add_or_merge_edge(G, *tids, edge_attrs)
            if show_inds == 'all':
                edge_labels[tuple(tids)] = ix
            elif show_inds == 'bond-size':
                edge_labels[tuple(tids)] = tn.ind_size(ix)

    # color the nodes
    colors = get_colors(color, custom_colors)

    # set parameters for all the nodes
    for tid, t in tn.tensor_map.items():

        if tid not in G.nodes:
            # e.g. tensor is a scalar
            if show_scalars:
                G.add_node(tid)
            else:
                continue

        G.nodes[tid]['size'] = node_size[tid]
        G.nodes[tid]['outline_size'] = node_outline_size[tid]
        color = node_color
        for tag in colors:
            if tag in t.tags:
                color = colors[tag]
        if tid in highlight_tids:
            color = highlight_tids_color
        G.nodes[tid]['color'] = color
        G.nodes[tid]['outline_color'] = tuple(
            (1.0 if i == 3 else node_outline_darkness) * c
            for i, c in enumerate(color)
        )
        G.nodes[tid]['marker'] = node_shape[tid]
        G.nodes[tid]['hatch'] = node_hatch[tid]

        if show_tags == 'tids':
            node_labels[tid] = str(tid)
        elif show_tags:
            # make the tags appear with auto vertical extent
            # node_label = '{' + str(list(t.tags))[1:-1] + '}'
            node_label = ', '.join(map(str, t.tags))
            node_labels[tid] = "\n".join(textwrap.wrap(
                node_label, max(2 * len(node_label) ** 0.5, 16)
            ))

    for hix in hyperedges:
        G.nodes[hix]['ind'] = hix
        G.nodes[hix]['color'] = (1.0, 1.0, 1.0, 1.0)
        G.nodes[hix]['size'] = 0.0
        G.nodes[hix]['outline_size'] = 0.0
        G.nodes[hix]['outline_color'] = (1.0, 1.0, 1.0, 1.0)
        G.nodes[hix]['marker'] = '.'  # set this to avoid warning - size is 0
        G.nodes[hix]['hatch'] = ''
        if show_inds == 'all':
            node_labels[hix] = hix
        elif show_inds == 'bond-size':
            node_labels[hix] = tn.ind_size(hix)

    if get == 'graph':
        return G

    if show_inds == 'bond-size':
        font_size = font_size_inner
        for oix in output_inds:
            node_labels[oix] = tn.ind_size(oix)
    elif show_inds:
        for oix in output_inds:
            node_labels[oix] = oix

    pos = get_positions(tn, G, fix, initial_layout, k, iterations,
                        use_forceatlas2, use_spring_weight)

    if get == 'pos':
        return pos

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
        fig.patch.set_alpha(0.0)
        ax.axis('off')
        ax.set_aspect('equal')
        if title is not None:
            ax.set_title(str(title))

        xmin = ymin = +float('inf')
        xmax = ymax = -float('inf')
        for xy in pos.values():
            xmin = min(xmin, xy[0])
            xmax = max(xmax, xy[0])
            ymin = min(ymin, xy[1])
            ymax = max(ymax, xy[1])

        if margin is None:
            # XXX: pad the plot range so that node circles are not clipped,
            #     using the networkx node_size parameter, *which is in absolute
            #     units* and so must be inverse transformed using matplotlib!
            inv = ax.transData.inverted()
            real_node_size = (abs(
                inv.transform((0, node_size.default_factory()))[1] -
                inv.transform((0, 0))[1]
            ) ** 0.5) / 4
            ax.set_xlim(xmin - real_node_size, xmax + real_node_size)
            ax.set_ylim(ymin - real_node_size, ymax + real_node_size)
        else:
            ax.margins(margin)
    else:
        fig = None

    nx.draw_networkx_edges(
        G, pos,
        width=tuple(x[2]['edge_size'] for x in G.edges(data=True)),
        edge_color=tuple(x[2]['color'] for x in G.edges(data=True)),
        alpha=edge_alpha,
        ax=ax,
    )

    # draw multiedges
    multiedge_centers = {}
    for i, j, attrs in G.edges(data=True):
        sizes = attrs['multiedge_sizes']
        multiplicity = len(sizes)
        if multiplicity > 1:
            rads = np.linspace(
                multiplicity * -multiedge_spread,
                multiplicity * +multiedge_spread,
                multiplicity
            )

            xa, ya = pos[i]
            xb, yb = pos[j]
            xab, yab = (xa + xb) / 2., (ya + yb) / 2.
            dx, dy = xb - xa, yb - ya

            inds = attrs['multiedge_inds']
            for sz, rad, ix in zip(sizes, rads, inds):

                # store the central point of the arc in case its needed by
                # the arrow drawing functionality
                cx, cy = xab + rad * dy * 0.5, yab - rad * dx * 0.5
                multiedge_centers[ix] = (cx, cy)

                ax.add_patch(patches.FancyArrowPatch(
                    (xa, ya), (xb, yb),
                    connectionstyle=patches.ConnectionStyle.Arc3(rad=rad),
                    alpha=edge_alpha,
                    linewidth=sz,
                    color=attrs['color'],
                    zorder=1,
                ))

    scatters = collections.defaultdict(lambda: collections.defaultdict(list))

    for node, attrs in G.nodes(data=True):
        # need to group by marker and hatch as matplotlib doesn't map these
        key = (attrs['marker'], attrs['hatch'])
        scatters[key]['x'].append(pos[node][0])
        scatters[key]['y'].append(pos[node][1])
        scatters[key]['s'].append(attrs['size'])
        scatters[key]['c'].append(attrs['color'])
        scatters[key]['linewidths'].append(attrs['outline_size'])
        scatters[key]['edgecolors'].append(attrs['outline_color'])

    # plot the nodes
    for (marker, hatch), data in scatters.items():
        ax.scatter(
            data['x'],
            data['y'],
            s=data['s'],
            c=data['c'],
            marker=marker,
            linewidths=data['linewidths'],
            edgecolors=data['edgecolors'],
            hatch=hatch,
            zorder=2,
            alpha=node_alpha,
        )

    # draw incomcing arrows for tensor left_inds
    if show_left_inds:
        for tid, t in tn.tensor_map.items():
            if t.left_inds is not None:
                for ind in t.left_inds:
                    if ind in hyperedges:
                        tida = ind
                    else:
                        tida, = (x for x in tn.ind_map[ind] if x != tid)
                    tidb = tid
                    (xa, ya), (xb, yb) = pos[tida], pos[tidb]

                    edge_width = G.get_edge_data(tida, tidb)['edge_size']
                    edge_length = ((xb - xa)**2 + (yb - ya)**2)**0.5
                    arrow_scale = (
                        0.02 * arrow_length * edge_width / edge_length**0.5
                    )

                    # arrow start and change
                    if ind in multiedge_centers:
                        x, y = multiedge_centers[ind]
                    else:
                        x = (xa + arrow_closeness * xb) / (1 + arrow_closeness)
                        y = (ya + arrow_closeness * yb) / (1 + arrow_closeness)

                    dx = (xb - xa) * arrow_scale
                    dy = (yb - ya) * arrow_scale

                    ax.add_patch(patches.FancyArrow(
                        x, y, dx, dy,
                        width=0,  # don't draw tail
                        length_includes_head=True,
                        head_width=(dx**2 + dy**2)**0.5,
                        head_length=(dx**2 + dy**2)**0.5,
                        linewidth=arrow_linewidth,
                        color=(
                            highlight_inds_color if ind in highlight_inds else
                            edge_color
                        ),
                        alpha=edge_alpha,
                        fill=True,
                        shape='full',
                        overhang=arrow_overhang,
                    ))

    if show_inds in {'all', 'bond-size'}:
        nx.draw_networkx_edge_labels(
            G, pos,
            edge_labels=edge_labels,
            font_size=font_size_inner,
            font_color=label_color,
            font_family=font_family,
            bbox={'ec': (0, 0, 0, 0), 'fc': (0, 0, 0, 0)},
            ax=ax,
        )
    if show_tags or show_inds:
        nx.draw_networkx_labels(
            G, pos,
            labels=node_labels,
            font_size=font_size,
            font_color=label_color,
            font_family=font_family,
            ax=ax,
        )

    # create legend
    if colors and legend:
        handles = []
        for color in colors.values():

            ecolor = tuple(
                (1.0 if i == 3 else node_outline_darkness) * c
                for i, c in enumerate(color)
            )
            linewidth = min(3, node_size.default_factory()**0.5 / 5)

            handles += [
                plt.Line2D(
                    [0], [0],
                    marker='o',
                    color=color,
                    markeredgecolor=ecolor,
                    markeredgewidth=linewidth,
                    linestyle='',
                    markersize=10
                )
            ]

        # needed in case '_' is the first character
        lbls = [f" {lbl}" for lbl in colors]

        legend = plt.legend(
            handles, lbls,
            ncol=max(round(len(handles) / 20), 1),
            loc='center left',
            bbox_to_anchor=(1, 0.5),
            labelcolor=label_color,
            prop={'family': font_family},
        )
        # do this manually as otherwise can't make only face transparent
        legend.get_frame().set_alpha(None)
        legend.get_frame().set_facecolor((0.0, 0.0, 0.0, 0.0))
        legend.get_frame().set_edgecolor((0.6, 0.6, 0.6, 0.2))

    if fig is None:
        # ax was supplied, don't modify and simply return
        return
    else:
        # axes and figure were created
        if xlims is not None:
            ax.set_xlim(xlims)
        if ylims is not None:
            ax.set_ylim(ylims)

    if return_fig:
        return fig
    else:
        plt.show()
        plt.close(fig)


# colorblind palettes by Bang Wong (https://www.nature.com/articles/nmeth.1618)

_COLORS_DEFAULT = (
    '#56B4E9',  # light blue
    '#E69F00',  # orange
    '#009E73',  # green
    '#D55E00',  # red
    '#F0E442',  # yellow
    '#CC79A7',  # purple
    '#0072B2',  # dark blue
)

_COLORS_SORTED = (
    '#0072B2',  # dark blue
    '#56B4E9',  # light blue
    '#009E73',  # green
    '#F0E442',  # yellow
    '#E69F00',  # orange
    '#D55E00',  # red
    '#CC79A7',  # purple
)


def mod_sat(c, mod):
    """Modify the luminosity of rgb color ``c``.
    """
    from matplotlib.colors import hsv_to_rgb, rgb_to_hsv

    h, s, v = rgb_to_hsv(c[:3])
    return (*hsv_to_rgb((h, mod * s, v)), 1.0)


def auto_colors(nc):
    import math
    from matplotlib.colors import LinearSegmentedColormap

    cmap = LinearSegmentedColormap.from_list('wong', _COLORS_SORTED)

    xs = list(map(cmap, np.linspace(0, 1.0, nc)))

    # modulate color saturation with sine to generate local distinguishability
    # ... but only turn on gradually for increasing number of nodes
    sat_mod_period = min(4, nc / 7)
    sat_mod_factor = max(0.0, 2 / 3 * math.tanh((nc - 7) / 4))

    return [
        mod_sat(
            c, 1 - sat_mod_factor * math.sin(math.pi * i / sat_mod_period)**2
        )
        for i, c in enumerate(xs)
    ]


def get_colors(color, custom_colors=None):
    """Generate a sequence of rgbs for tag(s) ``color``.
    """
    from matplotlib.colors import to_rgba

    if color is None:
        return dict()

    if isinstance(color, str):
        color = (color,)

    if custom_colors is not None:
        rgbs = list(map(to_rgba, custom_colors))
        return dict(zip(color, rgbs))

    nc = len(color)
    if nc <= 7:
        return dict(zip(color, list(map(to_rgba, _COLORS_DEFAULT))))

    rgbs = auto_colors(nc)
    return dict(zip(color, rgbs))


def _rotate(xy, theta):
    """Return a rotated set of points.
    """
    s = np.sin(theta)
    c = np.cos(theta)

    xyr = np.empty_like(xy)
    xyr[:, 0] = c * xy[:, 0] - s * xy[:, 1]
    xyr[:, 1] = s * xy[:, 0] + c * xy[:, 1]

    return xyr


def _span(xy):
    """Return the vertical span of the points.
    """
    return xy[:, 1].max() - xy[:, 1].min()


def _massage_pos(pos, nangles=360, flatten=False):
    """Rotate a position dict's points to cover a small vertical span
    """
    xy = np.empty((len(pos), 2))
    for i, (x, y) in enumerate(pos.values()):
        xy[i, 0] = x
        xy[i, 1] = y

    thetas = np.linspace(0, 2 * np.pi, nangles, endpoint=False)
    rxys = (_rotate(xy, theta) for theta in thetas)
    rxy0 = min(rxys, key=lambda rxy: _span(rxy))

    if flatten:
        rxy0[:, 1] /= 2

    return dict(zip(pos, rxy0))


def get_positions(
    tn,
    G,
    fix=None,
    initial_layout='auto',
    k=None,
    iterations='auto',
    use_forceatlas2=False,
    use_spring_weight=False,
):
    import networkx as nx

    if fix is None:
        fix = dict()
    else:
        fix = parse_dict_to_tids_or_inds(fix, tn)
        # find range with which to scale spectral points with
        xmin, xmax, ymin, ymax = (
            f(fix.values(), key=lambda xy: xy[i])[i]
            for f, i in [(min, 0), (max, 0), (min, 1), (max, 1)])
        if xmin == xmax:
            xmin, xmax = xmin - 1, xmax + 1
        if ymin == ymax:
            ymin, ymax = ymin - 1, ymax + 1
        xymin, xymax = min(xmin, ymin), max(xmax, ymax)

    if all(node in fix for node in G.nodes):
        # everything is already fixed
        return fix

    if initial_layout == 'auto':
        # automatically select
        if len(G) <= 100:
            # usually nicest
            initial_layout = 'kamada_kawai'
        else:
            # faster, but not as nice
            initial_layout = 'spectral'

    if iterations == 'auto':
        # the smaller the graph, the more iterations we can afford
        iterations = max(200, 1000 - len(G))

    # use spectral or other layout as starting point
    pos0 = getattr(nx, initial_layout + '_layout')(G)

    # scale points to fit with specified positions
    if fix:
        # but update with fixed positions
        pos0.update(valmap(lambda xy: np.array(
            (2 * (xy[0] - xymin) / (xymax - xymin) - 1,
             2 * (xy[1] - xymin) / (xymax - xymin) - 1)), fix))
        fixed = fix.keys()
    else:
        fixed = None

    # and then relax remaining using spring layout
    if iterations:

        if use_forceatlas2 is True:
            use_forceatlas2 = 1
        elif use_forceatlas2 in (0, False):
            use_forceatlas2 = float('inf')

        should_use_fa2 = (
            (fixed is None) and HAS_FA2 and (len(G) > use_forceatlas2)
        )

        weight = 'spring_weight' if use_spring_weight else None

        if should_use_fa2:
            from fa2 import ForceAtlas2
            # NB: some versions of fa2 don't support the `weight_attr` option
            pos = ForceAtlas2(verbose=False).forceatlas2_networkx_layout(
                G, pos=pos0, iterations=iterations)
        else:
            pos = nx.spring_layout(
                G, pos=pos0, fixed=fixed, k=k, iterations=iterations,
                weight=weight)
    else:
        pos = pos0

    if not fix:
        # finally rotate them to cover a small vertical span
        pos = _massage_pos(pos)

    return pos


def visualize_tensor(tensor, **kwargs):
    """Visualize all entries of a tensor, with indices mapped into the plane
    and values mapped into a color wheel.

    Parameters
    ----------
    tensor : Tensor
        The tensor to visualize.
    skew_factor : float, optional
        When there are more than two dimensions, a factor to scale the
        rotations by to avoid overlapping data points.
    size_map : bool, optional
        Whether to map the tensor value magnitudes to marker size.
    size_scale : float, optional
        An overall factor to scale the marker size by.
    alpha_map : bool, optional
        Whether to map the tensor value magnitudes to marker alpha.
    alpha_pow : float, optional
        The power to raise the magnitude to when mapping to alpha.
    alpha : float, optional
        The overall alpha to use for all markers if ``not alpha_map``.
    show_lattice : bool, optional
        Show a small grey dot for every 'lattice' point regardless of value.
    lattice_opts : dict, optional
        Options to pass to ``maplotlib.Axis.scatter`` for the lattice points.
    linewidths : float, optional
        The linewidth to use for the markers.
    marker : str, optional
        The marker to use for the markers.
    figsize : tuple, optional
        The size of the figure to create, if ``ax`` is not provided.
    ax : matplotlib.Axis, optional
        The axis to draw to. If not provided, a new figure will be created.

    Returns
    -------
    fig : matplotlib.Figure
        The figure containing the plot, or ``None`` if ``ax`` was provided.
    ax : matplotlib.Axis
        The axis containing the plot.
    """
    import xyzpy as xyz
    kwargs.setdefault('compass', True)
    kwargs.setdefault('compass_labels', tensor.inds)
    return xyz.visualize_tensor(tensor.data, **kwargs)



"""
Quantum Information for Many-Body calculations.
"""
import warnings

# some useful math
from math import pi, cos, sin, tan, exp, log, log2, log10, sqrt

# Core functions
from .core import (
    qarray,
    prod,
    isket,
    isbra,
    isop,
    isvec,
    issparse,
    isdense,
    isreal,
    isherm,
    ispos,
    mul,
    dag,
    dot,
    vdot,
    rdot,
    ldmul,
    rdmul,
    outer,
    explt,
    get_thread_pool,
    normalize,
    chop,
    quimbify,
    qu,
    ket,
    bra,
    dop,
    sparse,
    infer_size,
    trace,
    identity,
    eye,
    speye,
    dim_map,
    dim_compress,
    kron,
    kronpow,
    ikron,
    pkron,
    permute,
    itrace,
    partial_trace,
    expectation,
    expec,
    nmlz,
    tr,
    ptr,
)

# Linear algebra functions
from .linalg.base_linalg import (
    eigensystem,
    eig,
    eigh,
    eigvals,
    eigvalsh,
    eigvecs,
    eigvecsh,
    eigensystem_partial,
    groundstate,
    groundenergy,
    bound_spectrum,
    eigh_window,
    eigvalsh_window,
    eigvecsh_window,
    svd,
    svds,
    norm,
    expm,
    sqrtm,
    expm_multiply,
    Lazy,
)
from .linalg.rand_linalg import rsvd, estimate_rank
from .linalg.mpi_launcher import get_mpi_pool, can_use_mpi_pool

# Generating objects
from .gen.operators import (
    spin_operator,
    pauli,
    hadamard,
    phase_gate,
    S_gate,
    T_gate,
    U_gate,
    rotation,
    Rx,
    Ry,
    Rz,
    Xsqrt,
    Ysqrt,
    Zsqrt,
    Wsqrt,
    swap,
    iswap,
    fsim,
    fsimg,
    ncontrolled_gate,
    controlled,
    CNOT,
    cX,
    cY,
    cZ,
    ccX,
    ccY,
    ccZ,
    controlled_swap,
    fredkin,
    toffoli,
    ham_heis,
    ham_ising,
    ham_XY,
    ham_XXZ,
    ham_j1j2,
    ham_mbl,
    ham_heis_2D,
    zspin_projector,
    create,
    destroy,
    num,
    ham_hubbard_hardcore,
)
from .gen.states import (
    basis_vec,
    up,
    zplus,
    down,
    zminus,
    plus,
    xplus,
    minus,
    xminus,
    yplus,
    yminus,
    bloch_state,
    bell_state,
    singlet,
    thermal_state,
    neel_state,
    singlet_pairs,
    werner_state,
    ghz_state,
    w_state,
    levi_civita,
    perm_state,
    graph_state_1d,
    computational_state,
)
from .gen.rand import (
    randn,
    rand,
    rand_matrix,
    rand_herm,
    rand_pos,
    rand_rho,
    rand_ket,
    rand_uni,
    rand_haar_state,
    gen_rand_haar_states,
    rand_mix,
    rand_product_state,
    rand_matrix_product_state,
    rand_mps,
    rand_seperable,
    rand_iso,
    rand_mera,
    seed_rand,
    set_rand_bitgen,
)

# Functions for calculating properties
from .calc import (
    fidelity,
    purify,
    entropy,
    entropy_subsys,
    mutual_information,
    mutinf,
    mutinf_subsys,
    schmidt_gap,
    tr_sqrt,
    tr_sqrt_subsys,
    partial_transpose,
    negativity,
    logarithmic_negativity,
    logneg,
    logneg_subsys,
    concurrence,
    one_way_classical_information,
    quantum_discord,
    trace_distance,
    cprint,
    decomp,
    pauli_decomp,
    bell_decomp,
    correlation,
    pauli_correlations,
    ent_cross_matrix,
    qid,
    is_degenerate,
    is_eigenvector,
    page_entropy,
    heisenberg_energy,
    dephase,
    kraus_op,
    projector,
    measure,
    simulate_counts,
)

# Evolution class and methods
from .evo import Evolution

from .linalg.approx_spectral import (
    approx_spectral_function,
    tr_abs_approx,
    tr_exp_approx,
    tr_sqrt_approx,
    tr_xlogx_approx,
    entropy_subsys_approx,
    logneg_subsys_approx,
    negativity_subsys_approx,
    xlogx,
)
from .utils import (
    save_to_disk,
    load_from_disk,
    oset,
    LRU,
    tree_map,
    tree_apply,
    tree_flatten,
    tree_unflatten,
    format_number_with_error,
)


# versioneer
from . import _version

__version__ = _version.get_versions()["version"]


warnings.filterwarnings("ignore", message="Caching is not available when ")


__all__ = [
    # Accel ----------------------------------------------------------------- #
    "qarray",
    "prod",
    "isket",
    "isbra",
    "isop",
    "isvec",
    "issparse",
    "isdense",
    "isreal",
    "isherm",
    "ispos",
    "mul",
    "dag",
    "dot",
    "vdot",
    "rdot",
    "ldmul",
    "rdmul",
    "outer",
    "explt",
    # Core ------------------------------------------------------------------ #
    "normalize",
    "chop",
    "quimbify",
    "qu",
    "ket",
    "bra",
    "dop",
    "sparse",
    "infer_size",
    "trace",
    "identity",
    "eye",
    "speye",
    "dim_map",
    "dim_compress",
    "kron",
    "kronpow",
    "ikron",
    "pkron",
    "permute",
    "itrace",
    "partial_trace",
    "expectation",
    "expec",
    "nmlz",
    "tr",
    "ptr",
    # Linalg ---------------------------------------------------------------- #
    "eigensystem",
    "eig",
    "eigh",
    "eigvals",
    "eigvalsh",
    "eigvecs",
    "eigvecsh",
    "eigensystem_partial",
    "groundstate",
    "groundenergy",
    "bound_spectrum",
    "eigh_window",
    "eigvalsh_window",
    "eigvecsh_window",
    "svd",
    "svds",
    "norm",
    "Lazy",
    "rsvd",
    "estimate_rank",
    # Gen ------------------------------------------------------------------- #
    "spin_operator",
    "pauli",
    "hadamard",
    "phase_gate",
    "T_gate",
    "S_gate",
    "U_gate",
    "rotation",
    "Rx",
    "Ry",
    "Rz",
    "Xsqrt",
    "Ysqrt",
    "Zsqrt",
    "Wsqrt",
    "swap",
    "iswap",
    "fsim",
    "fsimg",
    "ncontrolled_gate",
    "controlled",
    "CNOT",
    "cX",
    "cY",
    "cZ",
    "ccX",
    "ccY",
    "ccZ",
    "controlled_swap",
    "fredkin",
    "toffoli",
    "ham_heis",
    "ham_ising",
    "ham_XY",
    "ham_XXZ",
    "ham_j1j2",
    "ham_mbl",
    "ham_heis_2D",
    "create",
    "destroy",
    "num",
    "ham_hubbard_hardcore",
    "zspin_projector",
    "basis_vec",
    "up",
    "zplus",
    "down",
    "zminus",
    "plus",
    "xplus",
    "minus",
    "xminus",
    "yplus",
    "yminus",
    "bloch_state",
    "bell_state",
    "singlet",
    "thermal_state",
    "neel_state",
    "singlet_pairs",
    "werner_state",
    "ghz_state",
    "w_state",
    "levi_civita",
    "perm_state",
    "graph_state_1d",
    "rand_matrix",
    "rand_herm",
    "rand_pos",
    "rand_rho",
    "rand_ket",
    "rand_uni",
    "rand_haar_state",
    "gen_rand_haar_states",
    "rand_mix",
    "rand_mps",
    "randn",
    "rand",
    "rand_product_state",
    "rand_matrix_product_state",
    "rand_seperable",
    "rand_iso",
    "rand_mera",
    "seed_rand",
    "set_rand_bitgen",
    "computational_state",
    # Calc ------------------------------------------------------------------ #
    "expm",
    "sqrtm",
    "expm_multiply",
    "fidelity",
    "purify",
    "entropy",
    "entropy_subsys",
    "mutual_information",
    "mutinf",
    "mutinf_subsys",
    "schmidt_gap",
    "tr_sqrt",
    "tr_sqrt_subsys",
    "partial_transpose",
    "negativity",
    "logarithmic_negativity",
    "logneg",
    "logneg_subsys",
    "concurrence",
    "one_way_classical_information",
    "quantum_discord",
    "trace_distance",
    "cprint",
    "decomp",
    "pauli_decomp",
    "bell_decomp",
    "correlation",
    "pauli_correlations",
    "ent_cross_matrix",
    "qid",
    "is_degenerate",
    "is_eigenvector",
    "page_entropy",
    "heisenberg_energy",
    "dephase",
    "kraus_op",
    "projector",
    "measure",
    "simulate_counts",
    # Evo ------------------------------------------------------------------- #
    "Evolution",
    # Approx spectral ------------------------------------------------------- #
    "approx_spectral_function",
    "tr_abs_approx",
    "tr_exp_approx",
    "tr_sqrt_approx",
    "tr_xlogx_approx",
    "entropy_subsys_approx",
    "logneg_subsys_approx",
    "negativity_subsys_approx",
    # Some misc useful math ------------------------------------------------- #
    "pi",
    "cos",
    "sin",
    "tan",
    "exp",
    "log",
    "log2",
    "log10",
    "sqrt",
    "xlogx",
    # Utils ----------------------------------------------------------------- #
    "save_to_disk",
    "load_from_disk",
    "get_thread_pool",
    "get_mpi_pool",
    "can_use_mpi_pool",
    "oset",
    "LRU",
    "tree_map",
    "tree_apply",
    "tree_flatten",
    "tree_unflatten",
]



from .contraction import (
    set_contract_path_cache,
    get_contract_strategy,
    set_contract_strategy,
    contract_strategy,
    get_contract_backend,
    set_contract_backend,
    contract_backend,
    get_tensor_linop_backend,
    set_tensor_linop_backend,
    tensor_linop_backend,
)
from .tensor_core import (
    tensor_contract,
    tensor_split,
    tensor_canonize_bond,
    tensor_compress_bond,
    tensor_direct_product,
    tensor_balance_bond,
    tensor_fuse_squeeze,
    tensor_network_sum,
    tensor_network_distance,
    tensor_network_fit_autodiff,
    tensor_network_fit_als,
    tensor_network_gate_inds,
    rand_uuid,
    bonds,
    bonds_size,
    group_inds,
    connect,
    new_bond,
    Tensor,
    TensorNetwork,
    PTensor,
    IsoTensor,
    COPY_tensor,
    oset,
)
from .interface import (
    pack,
    unpack,
    jax_register_pytree,
)
from .tensor_builder import (
    rand_tensor,
    rand_phased,
    TN_rand_reg,
    TN_from_sites_product_state,
    TN_from_sites_computational_state,
    TN2D_from_fill_fn,
    TN2D_empty,
    TN2D_with_value,
    TN2D_rand,
    TN3D_from_fill_fn,
    TN3D_empty,
    TN3D_with_value,
    TN3D_rand,
    TN2D_classical_ising_partition_function,
    TN3D_classical_ising_partition_function,
    HTN2D_classical_ising_partition_function,
    HTN3D_classical_ising_partition_function,
    HTN_classical_partition_function_from_edges,
    TN_classical_partition_function_from_edges,
    TN_dimer_covering_from_edges,
    TN_from_edges_and_fill_fn,
    TN_from_edges_empty,
    TN_from_edges_with_value,
    TN_from_edges_rand,
    TN_rand_from_edges,
    HTN_CP_from_sites_and_fill_fn,
    HTN_dual_from_edges_and_fill_fn,
    HTN_from_cnf,
    MPS_rand_state,
    MPS_product_state,
    MPS_computational_state,
    MPS_rand_computational_state,
    MPS_neel_state,
    MPS_ghz_state,
    MPS_w_state,
    MPS_zero_state,
    MPS_sampler,
    MPO_identity,
    MPO_identity_like,
    MPO_zeros,
    MPO_zeros_like,
    MPO_rand,
    MPO_rand_herm,
    MPO_product_operator,
    SpinHam,
    SpinHam1D,
    MPO_ham_ising,
    MPO_ham_XY,
    MPO_ham_heis,
    MPO_ham_mbl,
    ham_1d_ising,
    NNI_ham_ising,
    ham_1d_XY,
    NNI_ham_XY,
    ham_1d_heis,
    NNI_ham_heis,
    ham_1d_mbl,
    NNI_ham_mbl,
    ham_2d_ising,
    ham_2d_heis,
    ham_2d_j1j2,
    ham_3d_heis,
)
from .tensor_1d import (
    TensorNetwork1D,
    MatrixProductState,
    MatrixProductOperator,
    Dense1D,
    SuperOperator1D,
    align_TN_1D,
    expec_TN_1D,
    gate_TN_1D,
    superop_TN_1D,
    TNLinearOperator1D,
)
from .tensor_dmrg import (
    MovingEnvironment,
    DMRG,
    DMRG1,
    DMRG2,
    DMRGX,
)
from .tensor_mera import (
    MERA,
)
from .tensor_1d_tebd import (
    LocalHam1D,
    NNI,
    TEBD,
)
from .circuit import (
    Circuit,
    CircuitMPS,
    CircuitDense,
)
from .circuit_gen import (
    circ_ansatz_1D_zigzag,
    circ_ansatz_1D_brickwork,
    circ_ansatz_1D_rand,
    circ_qaoa,
)
from .optimize import (
    TNOptimizer,
)
from .tensor_arbgeom import (
    tensor_network_align,
    tensor_network_apply_op_vec,
)
from .tensor_arbgeom_tebd import (
    LocalHamGen,
    TEBDGen,
    SimpleUpdateGen,
)
from .tensor_2d import (
    gen_2d_bonds,
    TensorNetwork2D,
    PEPS,
    PEPO,
)
from .tensor_2d_tebd import (
    LocalHam2D,
    TEBD2D,
    SimpleUpdate,
    FullUpdate,
)
from .tensor_3d import (
    gen_3d_bonds,
    TensorNetwork3D,
    PEPS3D,
)
from .tensor_3d_tebd import (
    LocalHam3D,
)
from .geometry import (
    edges_2d_square,
    edges_2d_hexagonal,
    edges_2d_triangular,
    edges_2d_triangular_rectangular,
    edges_2d_kagome,
    edges_3d_cubic,
    edges_3d_diamond,
    edges_3d_diamond_cubic,
    edges_3d_pyrochlore,
)


__all__ = (
    "set_contract_path_cache",
    "contract_strategy",
    "get_contract_strategy",
    "set_contract_strategy",
    "contract_backend",
    "get_contract_backend",
    "set_contract_backend",
    "tensor_linop_backend",
    "get_tensor_linop_backend",
    "set_tensor_linop_backend",
    "tensor_contract",
    "tensor_split",
    "tensor_canonize_bond",
    "tensor_compress_bond",
    "tensor_direct_product",
    "tensor_balance_bond",
    "tensor_fuse_squeeze",
    "tensor_network_sum",
    "tensor_network_distance",
    "tensor_network_fit_autodiff",
    "tensor_network_fit_als",
    "tensor_network_gate_inds",
    "rand_uuid",
    "bonds",
    "bonds_size",
    "group_inds",
    "connect",
    "new_bond",
    "Tensor",
    "TensorNetwork",
    "TNLinearOperator1D",
    "PTensor",
    "IsoTensor",
    "COPY_tensor",
    "oset",
    "pack",
    "unpack",
    "jax_register_pytree",
    "rand_tensor",
    "rand_phased",
    "TN_rand_reg",
    "TN_from_sites_product_state",
    "TN_from_sites_computational_state",
    "TN2D_from_fill_fn",
    "TN2D_empty",
    "TN2D_with_value",
    "TN2D_rand",
    "TN3D_from_fill_fn",
    "TN3D_empty",
    "TN3D_with_value",
    "TN3D_rand",
    "TN2D_classical_ising_partition_function",
    "TN3D_classical_ising_partition_function",
    "HTN2D_classical_ising_partition_function",
    "HTN3D_classical_ising_partition_function",
    "HTN_classical_partition_function_from_edges",
    "TN_classical_partition_function_from_edges",
    "TN_dimer_covering_from_edges",
    "TN_from_edges_and_fill_fn",
    "TN_from_edges_empty",
    "TN_from_edges_with_value",
    "TN_from_edges_rand",
    "TN_rand_from_edges",
    "HTN_CP_from_sites_and_fill_fn",
    "HTN_dual_from_edges_and_fill_fn",
    "HTN_from_cnf",
    "MPS_rand_state",
    "MPS_product_state",
    "MPS_computational_state",
    "MPS_rand_computational_state",
    "MPS_neel_state",
    "MPS_ghz_state",
    "MPS_w_state",
    "MPS_zero_state",
    "MPS_sampler",
    "MPO_identity",
    "MPO_identity_like",
    "MPO_zeros",
    "MPO_zeros_like",
    "MPO_rand",
    "MPO_rand_herm",
    "MPO_product_operator",
    "SpinHam",
    "SpinHam1D",
    "MPO_ham_ising",
    "MPO_ham_XY",
    "MPO_ham_heis",
    "MPO_ham_mbl",
    "ham_1d_ising",
    "NNI_ham_ising",
    "ham_1d_XY",
    "NNI_ham_XY",
    "ham_1d_heis",
    "NNI_ham_heis",
    "ham_1d_mbl",
    "NNI_ham_mbl",
    "ham_2d_ising",
    "ham_2d_heis",
    "ham_2d_j1j2",
    "ham_3d_heis",
    "TensorNetwork1D",
    "MatrixProductState",
    "MatrixProductOperator",
    "Dense1D",
    "SuperOperator1D",
    "align_TN_1D",
    "expec_TN_1D",
    "gate_TN_1D",
    "superop_TN_1D",
    "MovingEnvironment",
    "DMRG",
    "DMRG1",
    "DMRG2",
    "DMRGX",
    "MERA",
    "TEBD",
    "LocalHam1D",
    "NNI",
    "Circuit",
    "CircuitMPS",
    "CircuitDense",
    "circ_ansatz_1D_zigzag",
    "circ_ansatz_1D_brickwork",
    "circ_ansatz_1D_rand",
    "circ_qaoa",
    "TNOptimizer",
    "tensor_network_align",
    "tensor_network_apply_op_vec",
    "LocalHamGen",
    "TEBDGen",
    "SimpleUpdateGen",
    "gen_2d_bonds",
    "TensorNetwork2D",
    "PEPS",
    "PEPO",
    "LocalHam2D",
    "TEBD2D",
    "SimpleUpdate",
    "FullUpdate",
    "gen_3d_bonds",
    "TensorNetwork3D",
    "PEPS3D",
    "LocalHam3D",
    "edges_2d_square",
    "edges_2d_hexagonal",
    "edges_2d_triangular",
    "edges_2d_triangular_rectangular",
    "edges_2d_kagome",
    "edges_3d_cubic",
    "edges_3d_diamond",
    "edges_3d_diamond_cubic",
    "edges_3d_pyrochlore",
)



import itertools

import pytest
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


class TestPEPSConstruct:

    @pytest.mark.parametrize('Lx', [3, 4, 5])
    @pytest.mark.parametrize('Ly', [3, 4, 5])
    def test_basic_rand(self, Lx, Ly):
        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=4)

        assert psi.max_bond() == 4
        assert psi.Lx == Lx
        assert psi.Ly == Ly
        assert len(psi.tensor_map) == Lx * Ly
        assert psi.site_inds == tuple(
            f'k{i},{j}' for i in range(Lx) for j in range(Ly)
        )
        assert psi.site_tags == tuple(
            f'I{i},{j}' for i in range(Lx) for j in range(Ly)
        )

        assert psi.bond_size((1, 1), (1, 2)) == (4)

        for i in range(Lx):
            assert len(psi.select(f'X{i}').tensor_map) == Ly
        for j in range(Ly):
            assert len(psi.select(f'Y{j}').tensor_map) == Lx

        for i in range(Lx):
            for j in range(Ly):
                assert psi.phys_dim(i, j) == 2
                assert isinstance(psi[i, j], qtn.Tensor)
                assert isinstance(psi[f'I{i},{j}'], qtn.Tensor)

        if Lx == Ly == 3:
            psi_dense = psi.to_dense(optimize='random-greedy')
            assert psi_dense.shape == (512, 1)

        psi.show()
        assert f'Lx={Lx}' in psi.__str__()
        assert f'Lx={Lx}' in psi.__repr__()

    def test_flatten(self):
        psi = qtn.PEPS.rand(3, 5, 3, seed=42)
        norm = psi.H & psi
        assert len(norm.tensors) == 30
        norm.flatten_()
        assert len(norm.tensors) == 15
        assert norm.max_bond() == 9

    def test_add_peps(self):
        pa = qtn.PEPS.rand(3, 4, 2)
        pb = qtn.PEPS.rand(3, 4, 3)
        pc = qtn.PEPS.rand(3, 4, 4)
        pab = pa + pb
        assert pab.max_bond() == 5
        assert pab @ pc == pytest.approx(pa @ pc + pb @ pc)

    @pytest.mark.parametrize('Lx', [3, 4, 5])
    @pytest.mark.parametrize('Ly', [3, 4, 5])
    def test_bond_coordinates(self, Lx, Ly):
        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=1)
        all_bonds = tuple(psi.gen_bond_coos())
        assert len(all_bonds) == 2 * Lx * Ly - Lx - Ly
        he = tuple(psi.gen_horizontal_even_bond_coos())
        ho = tuple(psi.gen_horizontal_odd_bond_coos())
        ve = tuple(psi.gen_vertical_even_bond_coos())
        vo = tuple(psi.gen_vertical_odd_bond_coos())
        for p in (he, ho, ve, vo):
            assert len(set(p)) == len(p)
            # check there is no overlap at all
            sites = tuple(itertools.chain.from_iterable(he))
            assert len(set(sites)) == len(sites)
        # check all coordinates are generated
        assert set(itertools.chain(he, ho, ve, vo)) == set(all_bonds)

    @pytest.mark.parametrize('where', [
        [(0, 0)], [(0, 1)], [(0, 2)], [(2, 2)],
        [(3, 2)], [(3, 1)], [(3, 0)], [(2, 0)], [(1, 1)],
    ])
    @pytest.mark.parametrize('contract', [False, True])
    def test_gate_2d_single_site(self, where, contract):
        Lx = 4
        Ly = 3
        D = 2

        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=D, seed=42, dtype=complex)
        psi_d = psi.to_dense()
        G = qu.rand_matrix(2)

        # compute the exact dense reference
        dims = [[2] * Ly] * Lx
        IGI = qu.ikron(G, dims, where, sparse=True)
        xe = (psi_d.H @ IGI @ psi_d).item()

        tn = psi.H & psi.gate(G, where, contract=contract)
        assert len(tn.tensors) == 2 * Lx * Ly + int(not contract)

        assert tn ^ all == pytest.approx(xe)

    @pytest.mark.parametrize(
        'contract', [False, True, 'split', 'reduce-split'])
    @pytest.mark.parametrize('where', [
        [(1, 1), (2, 1)], [(3, 2), (2, 2)],
        [(0, 0), (1, 1)], [(3, 1), (1, 2)]
    ])
    def test_gate_2d_two_site(self, where, contract):
        Lx = 4
        Ly = 3
        D = 2

        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=D, seed=42, dtype=complex)
        psi_d = psi.to_dense()

        # ikron can't tensor operators across non-adjacent subsytems
        # so we explicitly construct the gate as a sum of tensor components
        G_comps = [(qu.rand_matrix(2), qu.rand_matrix(2)) for _ in range(4)]
        G = sum(A & B for A, B in G_comps)

        # compute the exact dense reference
        dims = [[2] * Ly] * Lx
        IGI = sum(qu.ikron([A, B], dims, where, sparse=True)
                  for A, B in G_comps)

        xe = (psi_d.H @ IGI @ psi_d).item()

        tn = psi.H & psi.gate(G, where, contract=contract)
        change = {False: 1, True: -1, 'split': 0, 'reduce-split': 0}[contract]
        assert len(tn.tensors) == 2 * Lx * Ly + change

        assert tn ^ all == pytest.approx(xe)

    @pytest.mark.parametrize('propagate_tags',
                             [False, True, 'sites', 'register'])
    def test_gate_propagate_tags(self, propagate_tags):
        Lx = 4
        Ly = 3
        D = 1
        psi = qtn.PEPS.rand(Lx, Ly, D, tags='PSI0')
        psi.gate_(qu.rand_uni(4), [(1, 1), (1, 2)], tags='G1',
                  propagate_tags=propagate_tags)
        psi.gate_(qu.rand_uni(4), [(1, 2), (3, 2)], tags='G2',
                  propagate_tags=propagate_tags)
        if propagate_tags is False:
            assert set(psi['G1'].tags) == {'G1'}
            assert set(psi['G2'].tags) == {'G2'}
        if propagate_tags is True:
            tgs1 = {'I1,1', 'I1,2', 'G1', 'PSI0', 'Y1', 'Y2', 'X1'}
            assert set(psi['G1'][0].tags) == tgs1
            assert set(psi['G2'].tags) == tgs1 | {'G2', 'I3,2', 'X3', 'Y2'}
        if propagate_tags == 'sites':
            assert set(psi['G1'].tags) == {'G1', 'I1,1', 'I1,2'}
            assert set(psi['G2'].tags) == {'G2', 'I1,1', 'I1,2', 'I3,2'}
        if propagate_tags == 'register':
            assert set(psi['G1'].tags) == {'G1', 'I1,1', 'I1,2'}
            assert set(psi['G2'].tags) == {'G2', 'I1,2', 'I3,2'}


class Test2DContract:

    def test_contract_2d_one_layer_boundary(self):
        psi = qtn.PEPS.rand(4, 4, 3, seed=42)
        norm = psi.make_norm()
        xe = norm.contract(all, optimize='auto-hq')
        xt = norm.contract_boundary(max_bond=9)
        assert xt == pytest.approx(xe, rel=1e-2)

    def test_contract_2d_two_layer_boundary(self):
        psi = qtn.PEPS.rand(4, 4, 3, seed=42, tags='KET')
        norm = psi.make_norm()
        xe = norm.contract(all, optimize='auto-hq')
        xt = norm.contract_boundary(max_bond=27, layer_tags=['KET', 'BRA'])
        assert xt == pytest.approx(xe, rel=1e-2)

    def test_contract_2d_full_bond(self):
        psi = qtn.PEPS.rand(4, 4, 3, seed=42, tags='KET')
        norm = psi.make_norm()
        xe = norm.contract(all, optimize='auto-hq')
        xt = norm.contract_boundary(max_bond=27, mode='full-bond')
        assert xt == pytest.approx(xe, rel=1e-2)

    def test_ising_accuracy_regression(self):
        tn = qtn.TN2D_classical_ising_partition_function(16, 16, 0.44)
        Zex = 8.459419593253275e+100
        for s in 'bltr':
            Zap = tn.contract_boundary(max_bond=8, sequence=s)
            rerr = abs(1 - Zap / Zex)
            assert rerr < 2.2e-7
        for s in ['bt', 'lr']:
            Zap = tn.contract_boundary(max_bond=8, sequence=s)
            rerr = abs(1 - Zap / Zex)
            assert rerr < 3.9e-9

    @pytest.mark.parametrize("mode,two_layer", [
        ('mps', False),
        ('mps', True),
        ('full-bond', False),
    ])
    def test_compute_x_envs(self, mode, two_layer):
        psi = qtn.PEPS.rand(5, 4, 2, seed=42, tags='KET')
        norm = psi.make_norm()
        ex = norm.contract(all)

        if two_layer:
            compress_opts = {'cutoff': 1e-6, 'max_bond': 12, 'mode': mode,
                             'layer_tags': ['KET', 'BRA']}
        else:
            compress_opts = {'cutoff': 1e-6, 'max_bond': 8, 'mode': mode}
        row_envs = norm.compute_x_environments(**compress_opts)

        for i in range(norm.Lx):
            norm_i = (
                row_envs['xmin', i] &
                norm.select(norm.x_tag(i)) &
                row_envs['xmax', i]
            )
            x = norm_i.contract(all)
            assert x == pytest.approx(ex, rel=1e-2)

    @pytest.mark.parametrize("mode,two_layer", [
        ('mps', False),
        ('mps', True),
        ('full-bond', False),
    ])
    def test_compute_y_envs(self, mode, two_layer):
        psi = qtn.PEPS.rand(4, 5, 2, seed=42, tags='KET')
        norm = psi.retag({'KET': 'BRA'}).H | psi
        ex = norm.contract(all)

        if two_layer:
            compress_opts = {'cutoff': 1e-6, 'max_bond': 12, 'mode': mode,
                             'layer_tags': ['KET', 'BRA']}
        else:
            compress_opts = {'cutoff': 1e-6, 'max_bond': 8, 'mode': mode}
        col_envs = norm.compute_y_environments(**compress_opts)

        for j in range(norm.Lx):
            norm_j = (
                col_envs['ymin', j] &
                norm.select(norm.y_tag(j)) &
                col_envs['ymax', j]
            )
            x = norm_j.contract(all)
            assert x == pytest.approx(ex, rel=1e-2)

    def test_normalize(self):
        psi = qtn.PEPS.rand(4, 5, 2, seed=42)
        norm = (psi.H | psi).contract(all)
        assert norm != pytest.approx(1.0)
        psi.normalize_(balance_bonds=True, equalize_norms=True, cutoff=2e-3)
        norm = (psi.H | psi).contract(all)
        assert norm == pytest.approx(1.0, rel=0.01)

    @pytest.mark.parametrize('normalized', [False, True])
    @pytest.mark.parametrize('mode', ['mps', 'full-bond'])
    def test_compute_local_expectation_one_sites(self, mode, normalized):
        peps = qtn.PEPS.rand(4, 3, 2, seed=42, dtype='complex')

        # reference
        k = peps.to_dense()
        if normalized:
            qu.normalize(k)
        coos = list(itertools.product([0, 2, 3], [0, 1, 2]))
        terms = {coo: qu.rand_matrix(2) for coo in coos}
        dims = [[2] * 3] * 4
        A = sum(qu.ikron(A, dims, [coo], sparse=True)
                for coo, A in terms.items())
        ex = qu.expec(A, k)

        opts = dict(cutoff=2e-3, max_bond=9, contract_optimize='random-greedy')
        e = peps.compute_local_expectation(
            terms, mode=mode, normalized=normalized, **opts)

        assert e == pytest.approx(ex, rel=1e-2)

    @pytest.mark.parametrize('normalized', [False, True])
    @pytest.mark.parametrize('mode', ['mps', 'full-bond'])
    def test_compute_local_expectation_two_sites(self, mode, normalized):
        H = qu.ham_heis_2D(4, 3, sparse=True)
        Hij = qu.ham_heis(2, cyclic=False)

        peps = qtn.PEPS.rand(4, 3, 2, seed=42)
        k = peps.to_dense()

        if normalized:
            qu.normalize(k)
        ex = qu.expec(H, k)

        opts = dict(
            mode=mode,
            normalized=normalized,
            cutoff=2e-3,
            max_bond=16,
            contract_optimize='random-greedy'
        )

        # compute 2x1 and 1x2 plaquettes separately
        hterms = {coos: Hij for coos in peps.gen_horizontal_bond_coos()}
        vterms = {coos: Hij for coos in peps.gen_vertical_bond_coos()}

        he = peps.compute_local_expectation(hterms, **opts)
        ve = peps.compute_local_expectation(vterms, **opts)

        assert he + ve == pytest.approx(ex, rel=1e-2)

        # compute all terms in 2x2 plaquettes
        terms_all = {**hterms, **vterms}
        e = peps.compute_local_expectation(terms_all, autogroup=False, **opts)

        assert e == pytest.approx(ex, rel=1e-2)


class TestPEPO:

    @pytest.mark.parametrize('Lx', [3, 4, 5])
    @pytest.mark.parametrize('Ly', [3, 4, 5])
    def test_basic_rand(self, Lx, Ly):
        X = qtn.PEPO.rand_herm(Lx, Ly, bond_dim=4)

        assert X.max_bond() == 4
        assert X.Lx == Lx
        assert X.Ly == Ly
        assert len(X.tensor_map) == Lx * Ly
        assert X.upper_inds == tuple(
            f'k{i},{j}' for i in range(Lx) for j in range(Ly)
        )
        assert X.lower_inds == tuple(
            f'b{i},{j}' for i in range(Lx) for j in range(Ly)
        )
        assert X.site_tags == tuple(
            f'I{i},{j}' for i in range(Lx) for j in range(Ly)
        )

        assert X.bond_size((1, 1), (1, 2)) == (4)

        for i in range(Lx):
            assert len(X.select(f'X{i}').tensor_map) == Ly
        for j in range(Ly):
            assert len(X.select(f'Y{j}').tensor_map) == Lx

        for i in range(Lx):
            for j in range(Ly):
                assert X.phys_dim(i, j) == 2
                assert isinstance(X[i, j], qtn.Tensor)
                assert isinstance(X[f'I{i},{j}'], qtn.Tensor)

        if Lx == Ly == 3:
            X_dense = X.to_dense(optimize='random-greedy')
            assert X_dense.shape == (512, 512)
            assert qu.isherm(X_dense)

        X.show()
        assert f'Lx={Lx}' in X.__str__()
        assert f'Lx={Lx}' in X.__repr__()

    def test_add_pepo(self):
        pa = qtn.PEPO.rand(3, 4, 2)
        pb = qtn.PEPO.rand(3, 4, 3)
        pc = qtn.PEPO.rand(3, 4, 4)
        pab = pa + pb
        assert pab.max_bond() == 5
        assert pab @ pc == pytest.approx(pa @ pc + pb @ pc)

    def test_apply_pepo(self):
        A = qtn.PEPO.rand(Lx=3, Ly=2, bond_dim=2, seed=1)
        x = qtn.PEPS.rand(Lx=3, Ly=2, bond_dim=2, seed=0)
        y = A.apply(x)
        assert y.num_indices == x.num_indices
        Ad = A.to_dense()
        xd = x.to_dense()
        yd = y.to_dense()
        assert_allclose(Ad @ xd, yd)
        yc = A.apply(x, compress=True, max_bond=3)
        assert yc.max_bond() == 3


class TestMisc:

    def test_calc_plaquette_sizes(self):
        from quimb.tensor.tensor_2d import calc_plaquette_sizes
        H2 = {None: qu.ham_heis(2)}
        ham = qtn.LocalHam2D(10, 10, H2)
        assert calc_plaquette_sizes(ham.terms.keys()) == ((1, 2), (2, 1))
        assert (calc_plaquette_sizes(ham.terms.keys(), autogroup=False) ==
                ((2, 2),))
        H2[(1, 1), (2, 2)] = 0.5 * qu.ham_heis(2)
        ham = qtn.LocalHam2D(10, 10, H2)
        assert calc_plaquette_sizes(ham.terms.keys()) == ((2, 2),)
        H2[(2, 2), (2, 4)] = 0.25 * qu.ham_heis(2)
        H2[(2, 4), (4, 4)] = 0.25 * qu.ham_heis(2)
        ham = qtn.LocalHam2D(10, 10, H2)
        assert (calc_plaquette_sizes(ham.terms.keys()) ==
                ((1, 3), (2, 2), (3, 1)))
        assert (calc_plaquette_sizes(ham.terms.keys(), autogroup=False) ==
                ((3, 3),))

    def test_calc_plaquette_map(self):
        from quimb.tensor.tensor_2d import calc_plaquette_map
        plaquettes = [
            # 2x2 plaquette covering all sites
            ((0, 0), (2, 2)),
            # horizontal plaquettes
            ((0, 0), (1, 2)),
            ((1, 0), (1, 2)),
            # vertical plaquettes
            ((0, 0), (2, 1)),
            ((0, 1), (2, 1)),
        ]
        assert (
            calc_plaquette_map(plaquettes) ==
            {(0, 0): ((0, 0), (2, 1)),
             (0, 1): ((0, 1), (2, 1)),
             (1, 0): ((1, 0), (1, 2)),
             (1, 1): ((1, 0), (1, 2)),
             ((0, 0), (0, 1)): ((0, 0), (1, 2)),
             ((0, 0), (1, 0)): ((0, 0), (2, 1)),
             ((0, 0), (1, 1)): ((0, 0), (2, 2)),
             ((0, 1), (1, 0)): ((0, 0), (2, 2)),
             ((0, 1), (1, 1)): ((0, 1), (2, 1)),
             ((1, 0), (1, 1)): ((1, 0), (1, 2))}
        )



from math import log2
import itertools

import numpy as np

import quimb as qu
from .tensor_core import rand_uuid, IsoTensor, TensorNetwork
from .tensor_1d import TensorNetwork1D, TensorNetwork1DVector


def is_power_of_2(x):
    return ((x & (x - 1)) == 0) and x > 0


class MERA(TensorNetwork1DVector,
           TensorNetwork1D,
           TensorNetwork):
    r"""The Multi-scale Entanglement Renormalization Ansatz (MERA) state::

            ...     ...     ...     ...     ...     ...
         |       |       |       |       |       |       |
        ISO     ISO     ISO     ISO     ISO     ISO     ISO   :
           \   /   \   /   \   /   \   /   \   /   \   /      : '_LAYER1'
            UNI     UNI     UNI     UNI     UNI     UNI       :
           /   \   /   \   /   \   /   \   /   \   /   \
        O ISO ISO ISO ISO ISO ISO ISO ISO ISO ISO ISO ISO I   :
        | | | | | | | | | | | | | | | | | | | | | | | | | |   : '_LAYER0'
        UNI UNI UNI UNI UNI UNI UNI UNI UNI UNI UNI UNI UNI   :
        | | | | | | | | | | | | | | | | | | | | | | | | | |  <-- phys_dim
        0 1 2 3 4 ....                            ... L-2 L-1

    Parameters
    ----------
    L : int
        The number of phyiscal sites. Shoule be a power of 2.
    uni : array or sequence of arrays of shape (d, d, d, d).
        The unitary operator(s). These will be cycled over and placed from
        bottom left to top right in diagram above.
    iso : array or sequence of arrays of shape (d, d, d)
        The isometry operator(s). These will be cycled over and placed from
        bottom left to top right in diagram above.
    phys_dim : int, optional
        The dimension of the local hilbert space.
    dangle : bool, optional
        Whether to leave a dangling index on the final isometry, in order to
        maintain perfect scale invariance, else join the final unitaries just
        with an indentity.
    """

    _EXTRA_PROPS = ('_site_ind_id', '_site_tag_id', 'cyclic', '_L')
    _CONTRACT_STRUCTURED = False

    def __init__(self, L, uni=None, iso=None, phys_dim=2, dangle=False,
                 site_ind_id="k{}", site_tag_id="I{}", **tn_opts):

        # short-circuit for copying MERA
        if isinstance(L, MERA):
            super().__init__(L)
            for ep in MERA._EXTRA_PROPS:
                setattr(self, ep, getattr(L, ep))
            return

        self._site_ind_id = site_ind_id
        self._site_tag_id = site_tag_id
        self.cyclic = True
        self._L = L

        if not is_power_of_2(L):
            raise ValueError("``L`` should be a power of 2.")

        nlayers = round(log2(L))

        if hasattr(uni, 'shape'):
            uni = (uni,)

        if hasattr(iso, 'shape'):
            iso = (iso,)

        unis = itertools.cycle(uni)
        isos = itertools.cycle(iso)

        def gen_mera_tensors():
            u_ind_id = site_ind_id

            for i in range(nlayers):

                # index id connecting to layer below
                l_ind_id = u_ind_id
                # index id connecting to isos to unis
                m_ind_id = rand_uuid() + "_{}"
                # index id connecting to layer above
                u_ind_id = rand_uuid() + "_{}"

                # number of tensor sites in this layer
                eff_L = L // 2**i

                for j in range(0, eff_L, 2):

                    # generate the unitary:
                    #  ul | | ur
                    #     UNI
                    #  ll | | lr
                    #     j j+1
                    ll, lr = map(l_ind_id.format, (j, (j + 1) % eff_L))
                    ul, ur = map(m_ind_id.format, (j, (j + 1) % eff_L))
                    inds = (ll, lr, ul, ur)

                    tags = ("_UNI", f"_LAYER{i}")
                    if i == 0:
                        tags += (site_tag_id.format(j),
                                 site_tag_id.format(j + 1))

                    yield IsoTensor(next(unis), inds=inds,
                                    tags=tags, left_inds=(ll, lr))

                    # generate the isometry (offset by one effective site):
                    #      | ui
                    #     ISO
                    #  ll | | lr
                    #   j+1 j+2
                    ll, lr = map(m_ind_id.format, (j + 1, (j + 2) % eff_L))
                    ui = u_ind_id.format(j // 2)
                    inds = (ll, lr, ui)
                    tags = ("_ISO", f"_LAYER{i}")

                    if i < nlayers - 1 or dangle:
                        yield IsoTensor(next(isos), inds=inds,
                                        tags=tags, left_inds=(ll, lr))
                    else:
                        # don't leave dangling index at top
                        iso_f = next(isos)
                        yield IsoTensor(
                            np.eye(iso_f.shape[0], dtype=iso_f.dtype) / 2**0.5,
                            inds=inds[:-1], tags=tags, left_inds=(ll, lr)
                        )

        super().__init__(gen_mera_tensors(), virtual=True)

        # tag the MERA with the 'causal-cone' of each site
        for i in range(nlayers):
            for j in range(L):
                # get isometries in the same layer
                for t in self.select_neighbors(j):
                    if f'_LAYER{i}' in t.tags:
                        t.add_tag(f'I{j}')

                # get unitaries in layer above
                for t in self.select_neighbors(j):
                    if f'_LAYER{i + 1}' in t.tags:
                        t.add_tag(f'I{j}')

    @classmethod
    def rand(cls, L, max_bond=None, phys_dim=2, dtype=float, **mera_opts):

        d = phys_dim
        if max_bond is None:
            max_bond = d

        def gen_unis():
            D = d
            m = L // 2

            while True:
                for _ in range(m):
                    uni = qu.rand_iso(D**2, D**2, dtype=dtype)
                    uni.shape = (D, D, D, D)
                    yield uni

                D = min(D**2, max_bond)
                m //= 2

        def gen_isos():
            Dl = d
            Du = min(Dl**2, max_bond)
            m = L // 2

            while True:
                for _ in range(m):
                    iso = qu.rand_iso(Dl**2, Du, dtype=dtype)
                    iso.shape = (Dl, Dl, Du)
                    yield iso

                Dl = Du
                Du = min(Dl**2, max_bond)
                m //= 2

        return cls(L, gen_unis(), gen_isos(), phys_dim=d, **mera_opts)

    @classmethod
    def rand_invar(cls, L, phys_dim=2, dtype=float, **mera_opts):
        """Generate a random translational and scale invariant MERA.
        """
        d = phys_dim

        iso = qu.rand_iso(d**2, d, dtype=dtype)
        iso.shape = (d, d, d)

        uni = qu.rand_iso(d**2, d**2, dtype=dtype)
        uni.shape = (d, d, d, d)

        return cls(L, uni, iso, phys_dim=d, **mera_opts)



import pytest

import quimb as qu
import quimb.tensor as qtn


class Test3DManualContract:

    @pytest.mark.parametrize('canonize', [False, True])
    def test_ising_model(self, canonize):
        L = 5
        beta = 0.3
        fex = -2.7654417752878
        tn = qtn.TN3D_classical_ising_partition_function(L, L, L, beta=beta)
        Z = tn.contract_boundary(max_bond=8, canonize=canonize)
        f = -qu.log(Z) / (L**3 * beta)
        assert f == pytest.approx(fex, rel=1e-3)



import pytest
import numpy as np
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


class TestGeometries:

    @pytest.mark.parametrize('cyclic', [False, True])
    @pytest.mark.parametrize("edge_fn,shape,percell,coordination", [
        (qtn.edges_2d_square, (3, 3), 1, 4),
        (qtn.edges_2d_hexagonal, (3, 3), 2, 3),
        (qtn.edges_2d_kagome, (3, 3), 3, 4),
        (qtn.edges_2d_triangular, (3, 3), 1, 6),
        (qtn.edges_2d_triangular_rectangular, (3, 3), 2, 6),
        (qtn.edges_3d_cubic, (3, 3, 3), 1, 6),
        (qtn.edges_3d_pyrochlore, (3, 3, 3), 4, 6),
        (qtn.edges_3d_diamond, (3, 3, 3), 2, 4),
        (qtn.edges_3d_diamond_cubic, (2, 2, 2), 8, 4),
    ])
    def test_basic(self, cyclic, edge_fn, shape, percell, coordination):
        edges = edge_fn(*shape, cyclic=cyclic)
        tn = qtn.TN_rand_from_edges(edges, D=2)
        assert tn.num_tensors == qu.prod(shape) * percell
        assert max(t.ndim for t in tn) == coordination


class TestSpinHam1D:

    @pytest.mark.parametrize("cyclic", [False, True])
    def test_var_terms(self, cyclic):
        n = 8
        Hd = qu.ham_mbl(n, dh=0.77, seed=42, cyclic=cyclic)
        Ht = qtn.MPO_ham_mbl(n, dh=0.77, seed=42, cyclic=cyclic).to_dense()
        assert_allclose(Hd, Ht)

    @pytest.mark.parametrize("var_two", ['none', 'some', 'only'])
    @pytest.mark.parametrize("var_one", ['some', 'only', 'only-some',
                                         'def-only', 'none'])
    def test_specials(self, var_one, var_two):
        K1 = qu.rand_herm(2**1)

        n = 10
        HB = qtn.SpinHam1D(S=1 / 2)

        if var_two == 'some':
            HB += 1, K1, K1
            HB[4, 5] += 1, K1, K1
            HB[7, 8] += 1, K1, K1
        elif var_two == 'only':
            for i in range(n - 1):
                HB[i, i + 1] += 1, K1, K1
        else:
            HB += 1, K1, K1

        if var_one == 'some':
            HB += 1, K1
            HB[2] += 1, K1
            HB[3] += 1, K1
        elif var_one == 'only':
            for i in range(n - 1):
                HB[i] += 1, K1
        elif var_one == 'only-some':
            HB[1] += 1, K1
        elif var_one == 'def-only':
            HB += 1, K1

        HB.build_local_ham(n)
        H_mpo = HB.build_mpo(n)
        H_sps = HB.build_sparse(n)

        assert_allclose(H_mpo.to_dense(), H_sps.A)

    def test_no_default_term(self):
        N = 10
        builder = qtn.SpinHam1D(1 / 2)

        for i in range(N - 1):
            builder[i, i + 1] += 1.0, 'Z', 'Z'

        H = builder.build_mpo(N)

        dmrg = qtn.DMRG2(H)
        dmrg.solve(verbosity=1)

        assert dmrg.energy == pytest.approx(-2.25)


class TestMPSSpecificStates:

    def test_site_varying_phys_dim(self):
        k = qtn.MPS_rand_state(5, 4, phys_dim=[2, 3, 3, 2, 5])
        assert k.H @ k == pytest.approx(1.0)
        assert k.outer_dims_inds() == (
            (2, 'k0'), (3, 'k1'), (3, 'k2'), (2, 'k3'), (5, 'k4'),
        )

    @pytest.mark.parametrize("dtype", ['float32', 'complex64'])
    def test_ghz_state(self, dtype):
        mps = qtn.MPS_ghz_state(5, dtype=dtype)
        assert mps.dtype == dtype
        psi = qu.ghz_state(5, dtype=dtype)
        assert mps.H @ mps == pytest.approx(1.0)
        assert mps.bond_sizes() == [2, 2, 2, 2]
        assert qu.fidelity(psi, mps.to_dense()) == pytest.approx(1.0)

    @pytest.mark.parametrize("dtype", ['float32', 'complex64'])
    def test_w_state(self, dtype):
        mps = qtn.MPS_w_state(5, dtype=dtype)
        assert mps.dtype == dtype
        psi = qu.w_state(5, dtype=dtype)
        assert mps.H @ mps == pytest.approx(1.0)
        assert mps.bond_sizes() == [2, 2, 2, 2]
        assert qu.fidelity(psi, mps.to_dense()) == pytest.approx(1.0)

    def test_computational_state(self):
        mps = qtn.MPS_computational_state('01+-')
        assert_allclose(mps.to_dense(),
                        qu.up() & qu.down() & qu.plus() & qu.minus())


class TestMatrixProductOperatorSpecifics:

    def test_MPO_product_operator(self):
        psis = [qu.rand_ket(2) for _ in range(5)]
        ops = [qu.rand_matrix(2) for _ in range(5)]
        psif = qu.kron(*ops) @ qu.kron(*psis)
        mps = qtn.MPS_product_state(psis)
        mpo = qtn.MPO_product_operator(ops)
        assert mpo.bond_sizes() == [1, 1, 1, 1]
        mpsf = mpo.apply(mps)
        assert_allclose(mpsf.to_dense(), psif)


class TestGenericTN:

    def test_TN_rand_reg(self):
        n = 6
        reg = 3
        D = 2
        tn = qtn.TN_rand_reg(n, reg, D=D)
        assert tn.outer_inds() == ()
        assert tn.max_bond() == D
        assert {t.ndim for t in tn} == {reg}
        ket = qtn.TN_rand_reg(n, reg, D=2, phys_dim=2)
        assert set(ket.outer_inds()) == {f'k{i}' for i in range(n)}
        assert ket.max_bond() == D

    @pytest.mark.parametrize('Lx', [3])
    @pytest.mark.parametrize('Ly', [2, 4])
    @pytest.mark.parametrize('beta', [0.13, 0.44])
    @pytest.mark.parametrize('j', [-1.0, +1.0])
    @pytest.mark.parametrize('h', [0.0, 0.1])
    @pytest.mark.parametrize('cyclic',
                             [False, True, (False, True), (True, False)])
    def test_2D_classical_ising_model(self, Lx, Ly, beta, j, h, cyclic):
        tn = qtn.TN2D_classical_ising_partition_function(
            Lx, Ly, beta=beta, j=j, h=h, cyclic=cyclic)
        htn = qtn.HTN2D_classical_ising_partition_function(
            Lx, Ly, beta=beta, j=j, h=h, cyclic=cyclic)
        Z1 = tn.contract(all, output_inds=())
        Z2 = htn.contract(all, output_inds=())
        assert Z1 == pytest.approx(Z2)

        if not cyclic:
            # skip cyclic as nx has no multibonds for L=2
            import networkx as nx
            G = nx.lattice.grid_graph((Lx, Ly))
            Z3 = qtn.TN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z2 == pytest.approx(Z3)
            Z4 = qtn.HTN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z3 == pytest.approx(Z4)

    @pytest.mark.parametrize('Lx', [2])
    @pytest.mark.parametrize('Ly', [3])
    @pytest.mark.parametrize('Lz', [4])
    @pytest.mark.parametrize('beta', [0.13, 1 / 4.5])
    @pytest.mark.parametrize('j', [-1.0, +1.0])
    @pytest.mark.parametrize('h', [0.0, 0.1])
    @pytest.mark.parametrize('cyclic',
                             [False, True,
                              (False, True, False), (True, False, True)])
    def test_3D_classical_ising_model(self, Lx, Ly, Lz, beta, j, h, cyclic):
        tn = qtn.TN3D_classical_ising_partition_function(
            Lx, Ly, Lz, beta=beta, j=j, h=h, cyclic=cyclic)
        htn = qtn.HTN3D_classical_ising_partition_function(
            Lx, Ly, Lz, beta=beta, j=j, h=h, cyclic=cyclic)
        Z1 = tn.contract(all, output_inds=())
        Z2 = htn.contract(all, output_inds=())
        assert Z1 == pytest.approx(Z2)

        if not cyclic:
            # skip cyclic as nx has no multibonds for L=2
            import networkx as nx
            G = nx.lattice.grid_graph((Lx, Ly, Lz))
            Z3 = qtn.TN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z2 == pytest.approx(Z3)
            Z4 = qtn.HTN_classical_partition_function_from_edges(
                G.edges, beta=beta, j=j, h=h
            ).contract(all, output_inds=())
            assert Z3 == pytest.approx(Z4)

    def test_2d_classical_ising_varying_j(self):
        L = 5
        beta = 0.3
        edges = qtn.edges_2d_square(L, L)
        np.random.seed(666)
        js = {
            edge: np.random.normal()
            for edge in edges
        }
        tn = qtn.TN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x0 = tn.contract(all, output_inds=())
        tn = qtn.HTN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x1 = tn.contract(all, output_inds=())
        tn = qtn.TN2D_classical_ising_partition_function(
            L, L, beta=beta,  j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x2 = tn.contract(all, output_inds=())
        tn = qtn.HTN2D_classical_ising_partition_function(
            L, L, beta=beta,  j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x3 = tn.contract(all, output_inds=())
        assert x0 == pytest.approx(x1)
        assert x1 == pytest.approx(x2)
        assert x2 == pytest.approx(x3)

    def test_3d_classical_ising_varying_j(self):
        L = 3
        beta = 0.3
        edges = qtn.edges_3d_cubic(L, L, L)
        np.random.seed(666)
        js = {
            edge: np.random.normal()
            for edge in edges
        }
        tn = qtn.TN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x0 = tn.contract(all, output_inds=())
        tn = qtn.HTN_classical_partition_function_from_edges(
            edges, beta=beta, j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x1 = tn.contract(all, output_inds=())
        tn = qtn.TN3D_classical_ising_partition_function(
            L, L, L, beta=beta,  j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x2 = tn.contract(all, output_inds=())
        tn = qtn.HTN3D_classical_ising_partition_function(
            L, L, L, beta=beta,  j=lambda i, j: js[i, j])
        assert tn.dtype == 'float64'
        x3 = tn.contract(all, output_inds=())
        assert x0 == pytest.approx(x1)
        assert x1 == pytest.approx(x2)
        assert x2 == pytest.approx(x3)

    def test_tn_dimer_covering(self):
        edges = [(0, 1), (1, 2), (2, 3), (3, 0)]
        tn = qtn.TN_dimer_covering_from_edges(edges, cover_count=1)
        assert tn ^ all == pytest.approx(2.0)
        tn = qtn.TN_dimer_covering_from_edges(edges, cover_count=2)
        assert tn ^ all == pytest.approx(1.0)
        edges = [(0, 1), (1, 2), (2, 0)]
        tn = qtn.TN_dimer_covering_from_edges(edges, cover_count=1)
        assert tn ^ all == pytest.approx(0.0)

    def test_tn2d_fillers(self):
        tn = qtn.TN2D_empty(Lx=2, Ly=2, D=2)
        assert isinstance(tn, qtn.TensorNetwork2D)
        assert (
            (qtn.TN2D_rand(Lx=2, Ly=2, D=2, seed=42) ^ all) ==
            pytest.approx(qtn.TN2D_rand(Lx=2, Ly=2, D=2, seed=42) ^ all)
        )
        tn = qtn.TN2D_with_value(1.0, Lx=2, Ly=3, D=4)
        assert tn ^ all == pytest.approx(qu.prod(tn.ind_sizes().values()))

    def test_tn3d_fillers(self):
        tn = qtn.TN3D_empty(Lx=2, Ly=2, Lz=2, D=2)
        assert isinstance(tn, qtn.TensorNetwork3D)
        assert (
            (qtn.TN3D_rand(Lx=2, Ly=2, Lz=2, D=2, seed=42) ^ all) ==
            pytest.approx(qtn.TN3D_rand(Lx=2, Ly=2, Lz=2, D=2, seed=42) ^ all)
        )
        tn = qtn.TN3D_with_value(1.0, Lx=2, Ly=3, Lz=2, D=2)
        assert tn ^ all == pytest.approx(qu.prod(tn.ind_sizes().values()))



"""Backend agnostic array operations.
"""
import functools
import itertools

import numpy
from autoray import (
    compose,
    do,
    get_dtype_name,
    get_lib_fn,
    infer_backend,
    reshape,
)

from ..core import njit, qarray
from ..utils import compose as fn_compose
from ..linalg.base_linalg import norm_fro_dense


def asarray(array):
    """Maybe convert data for a tensor to use. If ``array`` already has a
    ``.shape`` attribute, i.e. looks like an array, it is left as-is. Else the
    elements are inspected to see which libraries' array constructor should be
    used, defaulting to ``numpy`` if everything is builtin or numpy numbers.
    """
    if isinstance(array, (numpy.matrix, qarray)):
        # if numpy make sure array not subclass
        return numpy.asarray(array)

    if hasattr(array, 'shape'):
        # otherwise don't touch things which are already array like
        return array

    # else we some kind of possibly nested python iterable -> inspect items
    backends = set()

    def _nd_py_iter(x):
        if isinstance(x, str):
            # handle recursion error
            return x

        backend = infer_backend(x)
        if backend != 'builtins':
            # don't iterate any non-builtin containers
            backends.add(backend)
            return x

        # is some kind of python container or element -> iterate or return
        try:
            return list(_nd_py_iter(sub) for sub in x)
        except TypeError:
            return x

    nested_tup = _nd_py_iter(array)

    # numpy and builtin elements treat as basic
    backends -= {"builtins", "numpy"}
    if not backends:
        backend = "numpy"
    else:
        backend, = backends

    return do('array', nested_tup, like=backend)


@functools.lru_cache(2**14)
def calc_fuse_perm_and_shape(shape, axes_groups):
    ndim = len(shape)

    # which group does each axis appear in, if any
    num_groups = len(axes_groups)
    ax2group = {ax: g for g, axes in enumerate(axes_groups) for ax in axes}

    # the permutation will be the same for every block: precalculate
    # n.b. all new groups will be inserted at the *first fused axis*
    position = min(g for gax in axes_groups for g in gax)
    axes_before = tuple(
        ax for ax in range(position)
        if ax2group.setdefault(ax, None) is None
    )
    axes_after = tuple(
        ax for ax in range(position, ndim)
        if ax2group.setdefault(ax, None) is None
    )
    perm = (*axes_before, *(ax for g in axes_groups for ax in g), *axes_after)

    # track where each axis will be in the new array
    new_axes = {ax: ax for ax in axes_before}
    for i, g in enumerate(axes_groups):
        for ax in g:
            new_axes[ax] = position + i
    for i, ax in enumerate(axes_after):
        new_axes[ax] = position + num_groups + i
    new_ndim = len(axes_before) + num_groups + len(axes_after)

    new_shape = [1] * new_ndim
    for i, d in enumerate(shape):
        g = ax2group[i]
        new_ax = new_axes[i]
        if g is None:
            # not fusing, new value is just copied
            new_shape[new_ax] = d
        else:
            # fusing: need to accumulate
            new_shape[new_ax] *= d

    return perm, tuple(new_shape)


@compose
def fuse(x, *axes_groups, backend=None):
    """Fuse the give group or groups of axes. The new fused axes will be
    inserted at the minimum index of any fused axis (even if it is not in
    the first group). For example, ``fuse(x, [5, 3], [7, 2, 6])`` will
    produce an array with axes like::

        groups inserted at axis 2, removed beyond that.
                ......<--
        (0, 1, g0, g1, 4, 8, ...)
                |   |
                |   g1=(7, 2, 6)
                g0=(5, 3)

    Parameters
    ----------
    axes_groups : sequence of sequences of int
        The axes to fuse. Each group of axes will be fused into a single
        axis.
    """
    if backend is None:
        backend = infer_backend(x)
    _transpose = get_lib_fn(backend, "transpose")
    _reshape = get_lib_fn(backend, "reshape")

    axes_groups =tuple(map(tuple, axes_groups))
    if not any(axes_groups):
        return x

    shape = tuple(map(int, x.shape))
    perm, new_shape = calc_fuse_perm_and_shape(shape, axes_groups)
    return _reshape(_transpose(x, perm), new_shape)


def ndim(array):
    """The number of dimensions of an array.
    """
    try:
        return array.ndim
    except AttributeError:
        return len(array.shape)


# ------------- miscelleneous other backend agnostic functions -------------- #

def iscomplex(x):
    """Does ``x`` have a complex dtype?
    """
    if infer_backend(x) == 'builtins':
        return isinstance(x, complex)
    return 'complex' in get_dtype_name(x)


@compose
def norm_fro(x):
    """The frobenius norm of an array.
    """
    try:
        return do("linalg.norm", reshape(x, (-1,)))
    except AttributeError:
        return do("sum", do("abs", x)**2) ** 0.5


norm_fro.register("numpy", norm_fro_dense)


def sensibly_scale(x):
    """Take an array and scale it *very* roughly such that random tensor
    networks consisting of such arrays do not have gigantic norms.
    """
    return x / norm_fro(x)**(1.5 / ndim(x))


@njit
def _numba_find_diag_axes(x, atol=1e-12):  # pragma: no cover
    """Numba-compiled array diagonal axis finder.

    Parameters
    ----------
    x : numpy.ndarray
        The array to search for diagonal axes.
    atol : float
        The tolerance with which to compare to zero.

    Returns
    -------
    diag_axes : set[tuple[int]]
        The set of pairs of axes which are diagonal.
    """

    # create the set of pairs of matching size axes
    diag_axes = set()
    for d1 in range(x.ndim - 1):
        for d2 in range(d1 + 1, x.ndim):
            if x.shape[d1] == x.shape[d2]:
                diag_axes.add((d1, d2))

    # enumerate through every array entry, eagerly invalidating axis pairs
    for index, val in numpy.ndenumerate(x):
        for d1, d2 in list(diag_axes):
            if (index[d1] != index[d2]) and (abs(val) > atol):
                diag_axes.remove((d1, d2))

        # all pairs invalid, nothing left to do
        if len(diag_axes) == 0:
            break

    return diag_axes


def find_diag_axes(x, atol=1e-12):
    """Try and find a pair of axes of ``x`` in which it is diagonal.

    Parameters
    ----------
    x : array-like
        The array to search.
    atol : float, optional
        Tolerance with which to compare to zero.

    Returns
    -------
    tuple[int] or None
        The two axes if found else None.

    Examples
    --------

        >>> x = np.array([[[1, 0], [0, 2]],
        ...               [[3, 0], [0, 4]]])
        >>> find_diag_axes(x)
        (1, 2)

    Which means we can reduce ``x`` without loss of information to:

        >>> np.einsum('abb->ab', x)
        array([[1, 2],
               [3, 4]])

    """
    shape = x.shape
    if len(shape) < 2:
        return None

    backend = infer_backend(x)

    # use numba-accelerated version for numpy arrays
    if backend == 'numpy':
        diag_axes = _numba_find_diag_axes(x, atol=atol)
        if diag_axes:
            # make it determinstic
            return min(diag_axes)
        return None
    indxrs = do('indices', shape, like=backend)

    for i, j in itertools.combinations(range(len(shape)), 2):
        if shape[i] != shape[j]:
            continue
        if do('allclose', x[indxrs[i] != indxrs[j]], 0.0,
              atol=atol, like=backend):
            return (i, j)
    return None


@njit
def _numba_find_antidiag_axes(x, atol=1e-12):  # pragma: no cover
    """Numba-compiled array antidiagonal axis finder.

    Parameters
    ----------
    x : numpy.ndarray
        The array to search for anti-diagonal axes.
    atol : float
        The tolerance with which to compare to zero.

    Returns
    -------
    antidiag_axes : set[tuple[int]]
        The set of pairs of axes which are anti-diagonal.
    """

    # create the set of pairs of matching size axes
    antidiag_axes = set()
    for i in range(x.ndim - 1):
        for j in range(i + 1, x.ndim):
            if x.shape[i] == x.shape[j]:
                antidiag_axes.add((i, j))

    # enumerate through every array entry, eagerly invalidating axis pairs
    for index, val in numpy.ndenumerate(x):
        for i, j in list(antidiag_axes):
            d = x.shape[i]
            if (index[i] != d - 1 - index[j]) and (abs(val) > atol):
                antidiag_axes.remove((i, j))

        # all pairs invalid, nothing left to do
        if len(antidiag_axes) == 0:
            break

    return antidiag_axes


def find_antidiag_axes(x, atol=1e-12):
    """Try and find a pair of axes of ``x`` in which it is anti-diagonal.

    Parameters
    ----------
    x : array-like
        The array to search.
    atol : float, optional
        Tolerance with which to compare to zero.

    Returns
    -------
    tuple[int] or None
        The two axes if found else None.

    Examples
    --------

        >>> x = np.array([[[0, 1], [0, 2]],
        ...               [[3, 0], [4, 0]]])
        >>> find_antidiag_axes(x)
        (0, 2)

    Which means we can reduce ``x`` without loss of information to:

        >>> np.einsum('aba->ab', x[::-1, :, :])
        array([[3, 4],
               [1, 2]])

    as long as we flip the order of dimensions on other tensors corresponding
    to the the same index.
    """
    shape = x.shape
    if len(shape) < 2:
        return None

    backend = infer_backend(x)

    # use numba-accelerated version for numpy arrays
    if backend == 'numpy':
        antidiag_axes = _numba_find_antidiag_axes(x, atol=atol)
        if antidiag_axes:
            # make it determinstic
            return min(antidiag_axes)
        return None

    indxrs = do('indices', shape, like=backend)

    for i, j in itertools.combinations(range(len(shape)), 2):
        di, dj = shape[i], shape[j]
        if di != dj:
            continue
        if do('allclose', x[indxrs[i] != dj - 1 - indxrs[j]], 0.0,
              atol=atol, like=backend):
            return (i, j)
    return None


@njit
def _numba_find_columns(x, atol=1e-12):  # pragma: no cover
    """Numba-compiled single non-zero column axis finder.

    Parameters
    ----------
    x :  array
        The array to search.
    atol : float, optional
        Absolute tolerance to compare to zero with.

    Returns
    -------
    set[tuple[int]]
        Set of pairs (axis, index) defining lone non-zero columns.
    """

    # possible pairings of axis + index
    column_pairs = set()
    for ax, d in enumerate(x.shape):
        for i in range(d):
            column_pairs.add((ax, i))

    # enumerate over all array entries, invalidating potential column pairs
    for index, val in numpy.ndenumerate(x):
        if abs(val) > atol:
            for ax, i in enumerate(index):
                for pax, pi in list(column_pairs):
                    if ax == pax and pi != i:
                        column_pairs.remove((pax, pi))

        # all potential pairs invalidated
        if not len(column_pairs):
            break

    return column_pairs


def find_columns(x, atol=1e-12):
    """Try and find columns of axes which are zero apart from a single index.

    Parameters
    ----------
    x : array-like
        The array to search.
    atol : float, optional
        Tolerance with which to compare to zero.

    Returns
    -------
    tuple[int] or None
        If found, the first integer is which axis, and the second is which
        column of that axis, else None.

    Examples
    --------

        >>> x = np.array([[[0, 1], [0, 2]],
        ...               [[0, 3], [0, 4]]])
        >>> find_columns(x)
        (2, 1)

    Which means we can happily slice ``x`` without loss of information to:

        >>> x[:, :, 1]
        array([[1, 2],
               [3, 4]])

    """
    shape = x.shape
    if len(shape) < 1:
        return None

    backend = infer_backend(x)

    # use numba-accelerated version for numpy arrays
    if backend == 'numpy':
        columns_pairs = _numba_find_columns(x, atol)
        if columns_pairs:
            return min(columns_pairs)
        return None

    indxrs = do('indices', shape, like=backend)

    for i in range(len(shape)):
        for j in range(shape[i]):
            if do('allclose', x[indxrs[i] != j], 0.0, atol=atol, like=backend):
                return (i, j)

    return None


class PArray:
    """Simple array-like object that lazily generates the actual array by
    calling a function with a set of parameters.

    Parameters
    ----------
    fn : callable
        The function that generates the tensor data from ``params``.
    params : sequence of numbers
        The initial parameters supplied to the generating function like
        ``fn(params)``.

    See Also
    --------
    PTensor
    """

    __slots__ = ('_fn', '_params', '_data', '_shape', '_shape_fn_id')

    def __init__(self, fn, params, shape=None):
        self.fn = fn
        self.params = params
        self._shape = shape
        self._shape_fn_id = id(fn)

    def copy(self):
        new = PArray(self.fn, self.params, self.shape)
        new._data = self._data  # for efficiency
        return new

    @property
    def fn(self):
        return self._fn

    @fn.setter
    def fn(self, x):
        self._fn = x
        self._data = None

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, x):
        self._params = asarray(x)
        self._data = None

    @property
    def data(self):
        if self._data is None:
            self._data = self._fn(self._params)
        return self._data

    @property
    def shape(self):
        # if we haven't calculated shape or have updated function, get shape
        _shape_fn_id = id(self.fn)
        if (self._shape is None) or (self._shape_fn_id != _shape_fn_id):
            self._shape = self.data.shape
            self._shape_fn_id = _shape_fn_id
        return self._shape

    @property
    def ndim(self):
        return len(self.shape)

    def add_function(self, g):
        """Chain the new function ``g`` on top of current function ``f`` like
        ``g(f(params))``.
        """
        f = self.fn
        self.fn = fn_compose(g, f)
